import json
import multiprocessing as mp
import os
import pickle
import queue
import signal
import threading
import time
from pathlib import Path

import pandas as pd
import pytest

from tabnetics.benchmarks import runner as benchmark


def _base_row(dataset_id: str, seed: int) -> dict:
    return {
        "dataset_id": dataset_id,
        "dataset_name": dataset_id,
        "tier": "easy",
        "effective_tier": "easy",
        "domain": "synthetic",
        "platform": "synthetic",
        "seed": seed,
        "config": "baseline",
        "protocol": "holdout",
        "data_source": "synthetic_enhanced",
        "validation_pipeline": "",
        "validation_scenario": "",
        "n_samples_total": 100,
        "n_features_total": 50,
        "n_train": 80,
        "n_test": 20,
        "n_fs_subset": 32,
        "accuracy": 0.9,
        "balanced_accuracy": 0.9,
        "macro_f1": 0.9,
        "hybrid_score": 0.9,
        "selected_features": 10,
        "model": "lr",
        "fs_time_sec": 1.0,
        "dist_time_sec": 1.0,
        "transform_time_sec": 1.0,
        "n_dist_features_fitted": 10,
        "n_dist_features_transformed": 8,
        "n_dist_rejected": 1,
        "n_dist_skipped_unreliable": 1,
        "n_dist_skipped_block_cv": 0,
        "n_low_gof_downweighted": 0,
        "mean_dist_stability_weight": 1.0,
        "cdf_block_gating_time_sec": 0.0,
        "cdf_block_gating_budget_hit": 0,
        "cdf_block_gating_blocks_evaluated": 0,
        "cdf_block_gating_blocks_applied": 0,
        "sota_holdout_bal_acc_low": 0.7,
        "sota_holdout_bal_acc_high": 1.0,
        "sota_inflated_bal_acc_low": 0.8,
        "sota_inflated_bal_acc_high": 1.0,
        "sota_holdout_status": "within",
        "sota_inflated_status": "within",
        "sota_bal_acc_low": 0.7,
        "sota_bal_acc_high": 1.0,
        "sota_status": "within",
        "protocol_gap_note": "",
        "sanity_ok": 1,
    }


def test_parallel_task_path_emits_outputs(tmp_path, monkeypatch):
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_easy_dfshift",
            "--seeds",
            "11",
            "23",
            "--ablation-profile",
            "none",
            "--max-workers",
            "2",
            "--output-dir",
            str(tmp_path),
        ]
    )

    captured = {"n_jobs": None}

    class FakeParallel:
        def __init__(self, n_jobs, prefer, verbose):
            captured["n_jobs"] = n_jobs

        def __call__(self, jobs):
            out = []
            for fn, fn_args, fn_kwargs in jobs:
                out.append(fn(*fn_args, **fn_kwargs))
            return out

    def fake_delayed(fn):
        def _wrapper(*fn_args, **fn_kwargs):
            return fn, fn_args, fn_kwargs

        return _wrapper

    def fake_task(ds_id, seed, _args):
        base = _base_row(ds_id, seed)
        nested = dict(base)
        nested["protocol"] = "nestedcv"
        nested["dataset_name"] = f"{ds_id} (NestedCV Audit)"
        nested["sota_holdout_status"] = "audit"
        nested["sota_inflated_status"] = "audit"
        nested["sota_status"] = "audit"
        bundle = {
            "schema_version": "1.0",
            "artifact_type": "df_fs_model_bundle",
            "dataset_id": ds_id,
            "seed": int(seed),
            "config": "baseline",
            "protocol": "holdout",
            "run_key": f"{ds_id}__seed{int(seed)}",
        }
        diag = {
            "schema_version": "1.0",
            "artifact_type": "df_fs_run_diagnostics",
            "dataset_id": ds_id,
            "seed": int(seed),
            "config": "baseline",
            "protocol": "holdout",
            "run_key": f"{ds_id}__seed{int(seed)}",
            "pipeline_stages": {"feature_selection": {}, "classifier_selection": {}},
        }
        return {
            "rows": [base, nested],
            "failures": [],
            "model_bundles": [bundle],
            "run_diagnostics": [diag],
        }

    monkeypatch.setattr(benchmark, "Parallel", FakeParallel)
    monkeypatch.setattr(benchmark, "delayed", fake_delayed)
    monkeypatch.setattr(benchmark, "_run_dataset_seed_task", fake_task)

    run_dir = benchmark.run_benchmark(args)

    runs_df = pd.read_csv(Path(run_dir) / "df_fs_runs.csv")
    metadata = json.loads((Path(run_dir) / "df_fs_metadata.json").read_text(encoding="utf-8"))
    model_bundles = json.loads((Path(run_dir) / "df_fs_model_bundles.json").read_text(encoding="utf-8"))
    run_diags = json.loads((Path(run_dir) / "df_fs_run_diagnostics.json").read_text(encoding="utf-8"))

    assert captured["n_jobs"] == 2
    assert set(runs_df["seed"].tolist()) == {11, 23}
    assert "domain" in runs_df.columns
    assert "platform" in runs_df.columns
    assert metadata["max_workers"] == 2
    assert "enable_model_cv_runtime_containment" in metadata["config_flags"]
    assert model_bundles["artifact_type"] == "df_fs_model_bundle_collection"
    assert run_diags["artifact_type"] == "df_fs_run_diagnostics_collection"
    assert "items" in model_bundles
    assert "items" in run_diags
    assert int(model_bundles.get("n_items", 0)) == 2
    assert int(run_diags.get("n_items", 0)) == 2


def test_pick_safe_nestedcv_splits_respects_min_train_per_class():
    y = [0] * 3 + [1] * 3
    n_splits, reason = benchmark._pick_safe_nestedcv_splits(
        y,
        requested_splits=5,
        min_train_per_class=2,
    )
    assert n_splits == 3
    assert reason == ""

    y_small = [0] * 2 + [1] * 2
    n_splits, reason = benchmark._pick_safe_nestedcv_splits(
        y_small,
        requested_splits=5,
        min_train_per_class=2,
    )
    assert n_splits is None
    assert "no_safe_n_splits" in reason


def test_task_timeout_context_raises_on_overrun():
    if not hasattr(signal, "SIGALRM"):
        pytest.skip("SIGALRM not supported on this platform")

    with pytest.raises(TimeoutError):
        with benchmark._task_timeout(0.05):
            time.sleep(0.2)


def test_hard_timeout_worker_kills_overrun_process(monkeypatch):
    if "fork" not in set(mp.get_all_start_methods()):
        pytest.skip("fork start method not available")

    def slow_worker(result_queue, cfg, X, y, dataset_name, seed, quiet_worker_logs):
        time.sleep(0.2)

    monkeypatch.setattr(benchmark, "_pipeline_run_worker", slow_worker)

    X = [[0.0, 1.0], [1.0, 0.0], [0.5, 0.5], [1.5, -0.2]]
    y = [0, 1, 0, 1]
    cfg = benchmark.DFFSConfig(
        random_seed=11,
        fs_fraction=0.5,
        n_final_features=2,
        max_dist_features=2,
        prefilter_top_k=2,
        enabled_methods=("mutual_information", "anova_f"),
    )

    with pytest.raises(TimeoutError):
        benchmark._run_pipeline_with_hard_timeout(
            cfg=cfg,
            X=X,
            y=y,
            dataset_name="tiny",
            seed=11,
            timeout_sec=0.05,
            quiet_worker_logs=True,
        )


def test_hard_timeout_worker_result_queue_no_false_empty(monkeypatch):
    if "fork" not in set(mp.get_all_start_methods()):
        pytest.skip("fork start method not available")

    def fast_worker(result_queue, cfg, X, y, dataset_name, seed, quiet_worker_logs):
        result_queue.put({"ok": True, "result": {"status": "ok", "seed": int(seed)}})

    monkeypatch.setattr(benchmark, "_pipeline_run_worker", fast_worker)

    import multiprocessing.queues as mpq

    original_empty = mpq.Queue.empty

    def _fail_if_empty_called(self):
        raise AssertionError("Queue.empty() must not be used in hard-timeout payload path")

    monkeypatch.setattr(mpq.Queue, "empty", _fail_if_empty_called)

    X = [[0.0, 1.0], [1.0, 0.0], [0.5, 0.5], [1.5, -0.2]]
    y = [0, 1, 0, 1]
    cfg = benchmark.DFFSConfig(
        random_seed=11,
        fs_fraction=0.5,
        n_final_features=2,
        max_dist_features=2,
        prefilter_top_k=2,
        enabled_methods=("mutual_information", "anova_f"),
    )

    result = benchmark._run_pipeline_with_hard_timeout(
        cfg=cfg,
        X=X,
        y=y,
        dataset_name="tiny",
        seed=11,
        timeout_sec=1.0,
        quiet_worker_logs=True,
    )
    assert result["status"] == "ok"
    assert result["seed"] == 11

    # Keep local reference to avoid lint false-positive if test body changes.
    assert callable(original_empty)


def test_hard_timeout_reads_result_queue_before_join(monkeypatch):
    monkeypatch.setattr(benchmark.mp, "get_all_start_methods", lambda: ["fork"])
    monkeypatch.setattr(
        benchmark.mp,
        "current_process",
        lambda: type("P", (), {"name": "MainProcess"})(),
    )

    state = {"got_payload": False}

    class _Queue:
        def __init__(self):
            self._item = None

        def put(self, item, timeout=None):
            self._item = item

        def get(self, timeout=None):
            state["got_payload"] = True
            if self._item is None:
                raise queue.Empty()
            return self._item

        def close(self):
            return None

        def join_thread(self):
            return None

    class _Process:
        def __init__(self, target, args):
            self._target = target
            self._args = args
            self._alive = True
            self.exitcode = 0
            self.daemon = False

        def start(self):
            self._target(*self._args)

        def join(self, timeout=None):
            assert state["got_payload"], "join() called before draining result_queue payload"
            self._alive = False

        def is_alive(self):
            return bool(self._alive)

        def terminate(self):
            self._alive = False

        def kill(self):
            self._alive = False

    class _Context:
        def Queue(self, maxsize=0):
            return _Queue()

        def Process(self, target, args):
            return _Process(target, args)

    monkeypatch.setattr(benchmark.mp, "get_context", lambda method: _Context())

    def fast_worker(result_queue, cfg, X, y, dataset_name, seed, quiet_worker_logs):
        result_queue.put({"ok": True, "result": {"status": "ok"}})

    monkeypatch.setattr(benchmark, "_pipeline_run_worker", fast_worker)

    cfg = benchmark.DFFSConfig(
        random_seed=11,
        fs_fraction=0.5,
        n_final_features=2,
        max_dist_features=2,
        prefilter_top_k=2,
        enabled_methods=("mutual_information", "anova_f"),
    )
    result = benchmark._run_pipeline_with_hard_timeout(
        cfg=cfg,
        X=[[0.0, 1.0], [1.0, 0.0], [0.5, 0.5], [1.5, -0.2]],
        y=[0, 1, 0, 1],
        dataset_name="tiny",
        seed=11,
        timeout_sec=1.0,
        quiet_worker_logs=True,
    )
    assert result["status"] == "ok"
    assert state["got_payload"] is True


def test_hard_timeout_loads_and_cleans_spilled_result_path(monkeypatch, tmp_path):
    if "fork" not in set(mp.get_all_start_methods()):
        pytest.skip("fork start method not available")

    payload_obj = {"status": "ok", "seed": 11, "metrics": {"ba": 0.82}}

    def spilled_worker(result_queue, cfg, X, y, dataset_name, seed, quiet_worker_logs):
        path = tmp_path / f"payload_{int(seed)}.pkl"
        with path.open("wb") as fh:
            pickle.dump(payload_obj, fh, protocol=pickle.HIGHEST_PROTOCOL)
        result_queue.put({"ok": True, "result_path": str(path)})

    monkeypatch.setattr(benchmark, "_pipeline_run_worker", spilled_worker)

    cfg = benchmark.DFFSConfig(
        random_seed=11,
        fs_fraction=0.5,
        n_final_features=2,
        max_dist_features=2,
        prefilter_top_k=2,
        enabled_methods=("mutual_information", "anova_f"),
    )
    payload_path = tmp_path / "payload_11.pkl"
    result = benchmark._run_pipeline_with_hard_timeout(
        cfg=cfg,
        X=[[0.0, 1.0], [1.0, 0.0], [0.5, 0.5], [1.5, -0.2]],
        y=[0, 1, 0, 1],
        dataset_name="tiny",
        seed=11,
        timeout_sec=1.0,
        quiet_worker_logs=True,
    )
    assert result == payload_obj
    assert payload_path.exists() is False


class _FakeQueue:
    def __init__(self):
        self._items = []

    def put(self, item, timeout=None):
        self._items.append(item)

    def put_nowait(self, item):
        self._items.append(item)

    def get(self, timeout=None):
        if not self._items:
            raise queue.Empty()
        return self._items.pop(0)

    def close(self):
        return None

    def join_thread(self):
        return None


class _FakeProcess:
    def __init__(self, target, args, *, start_exc=None):
        self._target = target
        self._args = args
        self._start_exc = start_exc
        self._alive = False
        self.exitcode = 0
        self.daemon = False

    def start(self):
        if self._start_exc is not None:
            raise self._start_exc
        self._alive = True
        self._target(*self._args)
        self._alive = False
        self.exitcode = 0

    def join(self, timeout=None):
        return None

    def is_alive(self):
        return bool(self._alive)

    def kill(self):
        self._alive = False
        self.exitcode = -9


class _FakeContext:
    def __init__(self, *, start_exc=None):
        self._start_exc = start_exc

    def Queue(self, maxsize=0):
        return _FakeQueue()

    def Process(self, target, args):
        return _FakeProcess(target, args, start_exc=self._start_exc)


def test_monitor_parallel_progress_emits_stall_watchdog_logs(monkeypatch, capsys):
    q = _FakeQueue()
    q.put(
        {
            "event": "task_start",
            "dataset_id": "synthetic_easy_dfshift",
            "seed": 11,
            "tier": "easy",
            "pid": 1234,
            "ts": time.time(),
        }
    )
    monkeypatch.setattr(
        benchmark,
        "_collect_descendant_wait_snapshot",
        lambda max_items=12: ["pid=1234 ppid=1 state=S wchan=futex_wait_queue_me cmd=loky-worker"],
    )

    stop_event = threading.Event()
    t = threading.Thread(
        target=benchmark._monitor_parallel_progress,
        args=(q, stop_event, 1, 5.0, 0.0, 0.05),
        daemon=True,
    )
    t.start()
    time.sleep(0.12)
    stop_event.set()
    t.join(timeout=2.0)

    out = capsys.readouterr().out
    assert "[stall-watchdog]" in out
    assert "no task completion" in out
    assert "futex_wait_queue_me" in out


def test_hard_timeout_prefers_spawn_inside_loky_worker(monkeypatch):
    monkeypatch.delenv("TABNETICS_HARD_TIMEOUT_START_METHOD", raising=False)
    monkeypatch.setattr(benchmark.mp, "get_all_start_methods", lambda: ["fork", "spawn"])
    monkeypatch.setattr(
        benchmark.mp,
        "current_process",
        lambda: type("P", (), {"name": "LokyProcess-5"})(),
    )
    context_calls = []

    def _get_context(method):
        context_calls.append(str(method))
        return _FakeContext()

    monkeypatch.setattr(benchmark.mp, "get_context", _get_context)

    def fast_worker(result_queue, cfg, X, y, dataset_name, seed, quiet_worker_logs):
        result_queue.put({"ok": True, "result": {"status": "ok", "seed": int(seed)}})

    monkeypatch.setattr(benchmark, "_pipeline_run_worker", fast_worker)

    cfg = benchmark.DFFSConfig(
        random_seed=11,
        fs_fraction=0.5,
        n_final_features=2,
        max_dist_features=2,
        prefilter_top_k=2,
        enabled_methods=("mutual_information", "anova_f"),
    )
    result = benchmark._run_pipeline_with_hard_timeout(
        cfg=cfg,
        X=[[0.0, 1.0], [1.0, 0.0], [0.5, 0.5], [1.5, -0.2]],
        y=[0, 1, 0, 1],
        dataset_name="tiny",
        seed=11,
        timeout_sec=1.0,
        quiet_worker_logs=True,
    )
    assert result["status"] == "ok"
    assert result["seed"] == 11
    assert context_calls
    assert context_calls[0] == "spawn"


def test_hard_timeout_spawn_launch_failure_falls_back_to_fork(monkeypatch):
    monkeypatch.delenv("TABNETICS_HARD_TIMEOUT_START_METHOD", raising=False)
    monkeypatch.setattr(benchmark.mp, "get_all_start_methods", lambda: ["fork", "spawn"])
    monkeypatch.setattr(
        benchmark.mp,
        "current_process",
        lambda: type("P", (), {"name": "LokyProcess-7"})(),
    )
    context_calls = []

    def _get_context(method):
        m = str(method)
        context_calls.append(m)
        if m == "spawn":
            return _FakeContext(start_exc=RuntimeError("spawn launch failure"))
        if m == "fork":
            return _FakeContext()
        raise AssertionError(f"Unexpected start method: {m}")

    monkeypatch.setattr(benchmark.mp, "get_context", _get_context)

    def fast_worker(result_queue, cfg, X, y, dataset_name, seed, quiet_worker_logs):
        result_queue.put({"ok": True, "result": {"status": "ok"}})

    monkeypatch.setattr(benchmark, "_pipeline_run_worker", fast_worker)

    cfg = benchmark.DFFSConfig(
        random_seed=11,
        fs_fraction=0.5,
        n_final_features=2,
        max_dist_features=2,
        prefilter_top_k=2,
        enabled_methods=("mutual_information", "anova_f"),
    )
    with pytest.warns(RuntimeWarning, match="falling back to 'fork'"):
        result = benchmark._run_pipeline_with_hard_timeout(
            cfg=cfg,
            X=[[0.0, 1.0], [1.0, 0.0], [0.5, 0.5], [1.5, -0.2]],
            y=[0, 1, 0, 1],
            dataset_name="tiny",
            seed=11,
            timeout_sec=1.0,
            quiet_worker_logs=True,
        )
    assert result["status"] == "ok"
    assert context_calls[:2] == ["spawn", "fork"]


def test_subspace_method_set_and_ablation_toggle_present():
    assert "mnpo_subspace_extended" in benchmark.FS_METHOD_SETS
    assert "mnpo_copula_extended" in benchmark.FS_METHOD_SETS
    assert "mnpo_tigress_extended" in benchmark.FS_METHOD_SETS
    assert "mnpo_rankagg_extended" in benchmark.FS_METHOD_SETS
    assert "mnpo_ova_extended" in benchmark.FS_METHOD_SETS
    assert "mnpo_ecoc_class_aware_extended" in benchmark.FS_METHOD_SETS
    assert "mnpo_iterative_pruning_extended" in benchmark.FS_METHOD_SETS
    assert "mnpo_joint_multiclass_extended" in benchmark.FS_METHOD_SETS
    assert "mnpo_dove_extended" in benchmark.FS_METHOD_SETS
    assert "mnpo_sparse_multinomial_extended" in benchmark.FS_METHOD_SETS
    assert "mnpo_nsc_extended" in benchmark.FS_METHOD_SETS
    assert "a22_nsc" in benchmark.FS_METHOD_SETS
    assert "mnpo_nsc_threshold_variants_extended" in benchmark.FS_METHOD_SETS
    assert "mnpo_class_pareto_extended" in benchmark.FS_METHOD_SETS
    assert "a28_class_pareto" in benchmark.FS_METHOD_SETS
    assert "mnpo_hsic_lasso_extended" in benchmark.FS_METHOD_SETS
    assert "a25_hsic_lasso" in benchmark.FS_METHOD_SETS
    assert "mnpo_slce_extended" in benchmark.FS_METHOD_SETS
    assert "a29_slce" in benchmark.FS_METHOD_SETS
    assert "mnpo_dove_sparse_multinomial_extended" in benchmark.FS_METHOD_SETS
    assert "mnpo_iterative_pruning_bounded_extended" in benchmark.FS_METHOD_SETS
    assert "mnpo_iterative_pruning_bounded_cpss_extended" in benchmark.FS_METHOD_SETS
    assert "mnpo_iterative_pruning_bounded_pareto_extended" in benchmark.FS_METHOD_SETS
    assert "mnpo_iterative_pruning_bounded_pareto_stability_extended" in benchmark.FS_METHOD_SETS
    assert "mnpo_broad_stable" in benchmark.FS_METHOD_SETS


def test_build_base_config_wires_classifier_hybrid_flags():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_easy_dfshift",
            "--seeds",
            "11",
            "--ablation-profile",
            "none",
            "--classifier-selection-mode",
            "mnpo_hybrid",
            "--classification-backend",
            "optuna",
            "--optuna-time-budget",
            "45",
            "--optuna-n-trials",
            "7",
            "--classifier-oracle-k",
            "2",
            "--classifier-oracle-weighting-mode",
            "tritrust",
            "--include-nsc-model",
            "--include-pls-da-model",
            "--include-gpc-model",
            "--include-lgbm-model",
            "--include-extra-tree-model",
            "--include-catboost-model",
            "--fs-use-conformal-uq",
            "--fs-conformal-uq-alpha",
            "0.12",
            "--fs-conformal-uq-min-folds",
            "6",
            "--enable-stage2-ratio-augmentation",
            "--stage2-ratio-max-features",
            "12",
            "--stage2-ratio-selection-method",
            "correlation",
            "--enable-classifier-conformal",
            "--classifier-conformal-alpha",
            "0.13",
            "--classifier-conformal-calibration-fraction",
            "0.30",
            "--classifier-conformal-min-calibration",
            "12",
            "--classifier-conformal-output-sets",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_easy_dfshift"]
    cfg = benchmark._build_base_config(args, spec, seed=11)

    assert cfg.classification.selection_mode == "mnpo_hybrid"
    assert cfg.classification.oracle_k == 2
    assert cfg.classification.oracle_weighting_mode == "tritrust"
    assert cfg.classification.include_nsc_model is True
    assert cfg.classification.include_pls_da_model is True
    assert cfg.classification.include_gpc_model is True
    assert cfg.classification.include_lgbm_model is True
    assert cfg.classification.include_extra_tree_model is True
    assert cfg.classification.include_catboost_model is True
    assert cfg.classification.backend == "optuna"
    assert cfg.classification.optuna_time_budget == 45
    assert cfg.classification.optuna_n_trials == 7
    assert bool(cfg.fs_use_conformal_uq) is True
    assert float(cfg.fs_conformal_uq_alpha) == pytest.approx(0.12)
    assert int(cfg.fs_conformal_uq_min_folds) == 6
    assert cfg.classification.stage2_ratio_augmentation_enabled is True
    assert cfg.classification.stage2_ratio_max_features == 12
    assert cfg.classification.stage2_ratio_selection_method == "correlation"
    assert cfg.classification.conformal_enabled is True
    assert cfg.classification.conformal_alpha == pytest.approx(0.13)
    assert cfg.classification.conformal_calibration_fraction == pytest.approx(0.30)
    assert cfg.classification.conformal_min_calibration == 12
    assert cfg.classification.conformal_output_sets is True
    assert "mnpo_broad_all" in benchmark.FS_METHOD_SETS
    assert "mnpo_broad_bundle_a" in benchmark.FS_METHOD_SETS
    assert "mnpo_broad_bundle_b" in benchmark.FS_METHOD_SETS
    assert "mnpo_broad_bundle_c" in benchmark.FS_METHOD_SETS

    base = benchmark.DFFSConfig(
        enabled_methods=("subspace_stability", "linear_svm", "mutual_information", "anova_f")
    )
    cfgs = benchmark._build_ablation_configs(base, profile="core")
    names = [name for name, _ in cfgs]

    assert "fs_no_subspace_stability" in names


def test_broad_profiles_have_expected_method_coverage():
    stable = benchmark.FS_METHOD_SETS["mnpo_broad_stable"]
    broad_all = benchmark.FS_METHOD_SETS["mnpo_broad_all"]

    assert len(stable) == 14
    assert "class_pareto_front" in stable
    assert "copula_knockoff" in stable
    assert "relieff" in stable
    assert "hsic_lasso" in stable

    assert len(broad_all) >= 25
    assert "iterative_redundancy_pruning_bounded" in broad_all
    assert "joint_auc_l1" in broad_all
    assert "ktsp" in broad_all
    assert "fcbf" in broad_all
    assert "cmim" in broad_all
    assert "treeshap" in broad_all
    assert "oaenet" in broad_all


def test_broad_bundle_overlays_apply_expected_toggles():
    parser = benchmark.build_arg_parser()
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]

    args_a = parser.parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--fs-method-set",
            "mnpo_broad_bundle_a",
        ]
    )
    cfg_a = benchmark._build_base_config(args_a, spec, seed=11)
    assert bool(cfg_a.fs_adaptive_portfolio_sizing_enabled) is True
    assert bool(cfg_a.use_qre_smoothing) is True
    assert bool(cfg_a.use_oracle_redundancy_penalty) is True

    args_b = parser.parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--fs-method-set",
            "mnpo_broad_bundle_b",
        ]
    )
    cfg_b = benchmark._build_base_config(args_b, spec, seed=11)
    assert bool(cfg_b.fs_wrapper_refine_enabled) is True
    assert bool(cfg_b.fs_rashomon_enabled) is True

    args_c = parser.parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--fs-method-set",
            "mnpo_broad_bundle_c",
        ]
    )
    cfg_c = benchmark._build_base_config(args_c, spec, seed=11)
    assert bool(cfg_c.screening_enabled) is True
    assert str(cfg_c.screening_method) == "stir"


def test_shadow_evaluator_helper_concordance_table():
    rows = [
        {
            **_base_row("synthetic_easy_dfshift", 11),
            "protocol": "holdout",
            "sota_holdout_status": "within",
            "balanced_accuracy": 0.92,
            "hybrid_score": 0.91,
        },
        {
            **_base_row("synthetic_easy_dfshift", 23),
            "protocol": "holdout",
            "sota_holdout_status": "within",
            "balanced_accuracy": 0.93,
            "hybrid_score": 0.20,
        },
    ]
    shadow_df, meta = benchmark._build_shadow_evaluator_pilot(
        rows,
        frozen_dataset_ids=["synthetic_easy_dfshift"],
    )
    assert not shadow_df.empty
    assert int(meta["n_compared"]) == 2
    assert int(meta["disagreement_count"]) >= 1


def test_parallel_task_path_emits_shadow_artifact(tmp_path, monkeypatch):
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_easy_dfshift",
            "--seeds",
            "11",
            "--ablation-profile",
            "none",
            "--max-workers",
            "1",
            "--enable-shadow-evaluator",
            "--shadow-frozen-datasets",
            "synthetic_easy_dfshift",
            "--output-dir",
            str(tmp_path),
        ]
    )

    def fake_task(ds_id, seed, _args):
        base = _base_row(ds_id, seed)
        base["hybrid_score"] = 0.95
        return {"rows": [base], "failures": []}

    monkeypatch.setattr(benchmark, "_run_dataset_seed_task", fake_task)

    run_dir = benchmark.run_benchmark(args)
    metadata = json.loads((Path(run_dir) / "df_fs_metadata.json").read_text(encoding="utf-8"))
    shadow_path = Path(run_dir) / "shadow_evaluator_pilot.csv"

    assert metadata["shadow_evaluator"]["enabled"] is True
    assert shadow_path.exists()


def test_rank_aggregation_ablation_toggle_present():
    base = benchmark.DFFSConfig(
        enabled_methods=("stability_subsample", "linear_svm", "mutual_information", "anova_f"),
        fs_rank_aggregation_mode="rra",
    )
    cfgs = benchmark._build_ablation_configs(base, profile="core")
    names = [name for name, _ in cfgs]

    assert "fs_disable_rank_aggregation" in names


def test_wrapper_refine_ablation_toggle_present():
    base = benchmark.DFFSConfig(
        enabled_methods=("stability_subsample", "linear_svm", "mutual_information", "anova_f"),
        fs_wrapper_refine_enabled=True,
    )
    cfgs = benchmark._build_ablation_configs(base, profile="core")
    names = [name for name, _ in cfgs]

    assert "fs_disable_wrapper_refine" in names


def test_ova_ablation_toggle_present():
    base = benchmark.DFFSConfig(
        enabled_methods=("ova_ensemble", "linear_svm", "mutual_information", "anova_f"),
    )
    cfgs = benchmark._build_ablation_configs(base, profile="core")
    names = [name for name, _ in cfgs]

    assert "fs_no_ova_ensemble" in names


def test_ecoc_ablation_toggle_present():
    base = benchmark.DFFSConfig(
        enabled_methods=("ecoc_class_aware", "linear_svm", "mutual_information", "anova_f"),
    )
    cfgs = benchmark._build_ablation_configs(base, profile="core")
    names = [name for name, _ in cfgs]

    assert "fs_no_ecoc_class_aware" in names


def test_joint_multiclass_ablation_toggle_present():
    base = benchmark.DFFSConfig(
        enabled_methods=("joint_multiclass_support", "linear_svm", "mutual_information", "anova_f"),
    )
    cfgs = benchmark._build_ablation_configs(base, profile="core")
    names = [name for name, _ in cfgs]

    assert "fs_no_joint_multiclass_support" in names


def test_dove_ablation_toggle_present():
    base = benchmark.DFFSConfig(
        enabled_methods=("dove_class_specific", "linear_svm", "mutual_information", "anova_f"),
    )
    cfgs = benchmark._build_ablation_configs(base, profile="core")
    names = [name for name, _ in cfgs]

    assert "fs_no_dove_class_specific" in names


def test_sparse_multinomial_ablation_toggle_present():
    base = benchmark.DFFSConfig(
        enabled_methods=("sparse_multinomial", "linear_svm", "mutual_information", "anova_f"),
    )
    cfgs = benchmark._build_ablation_configs(base, profile="core")
    names = [name for name, _ in cfgs]

    assert "fs_no_sparse_multinomial" in names


def test_nsc_ablation_toggle_present():
    base = benchmark.DFFSConfig(
        enabled_methods=("nearest_shrunken_centroid", "linear_svm", "mutual_information", "anova_f"),
    )
    cfgs = benchmark._build_ablation_configs(base, profile="core")
    names = [name for name, _ in cfgs]

    assert "fs_no_nearest_shrunken_centroid" in names


def test_class_pareto_ablation_toggle_present():
    base = benchmark.DFFSConfig(
        enabled_methods=("class_pareto_front", "linear_svm", "mutual_information", "anova_f"),
    )
    cfgs = benchmark._build_ablation_configs(base, profile="core")
    names = [name for name, _ in cfgs]

    assert "fs_no_class_pareto_front" in names


def test_hsic_lasso_ablation_toggle_present():
    base = benchmark.DFFSConfig(
        enabled_methods=("hsic_lasso", "linear_svm", "mutual_information", "anova_f"),
    )
    cfgs = benchmark._build_ablation_configs(base, profile="core")
    names = [name for name, _ in cfgs]

    assert "fs_no_hsic_lasso" in names


def test_iterative_pruning_ablation_toggle_present():
    base = benchmark.DFFSConfig(
        enabled_methods=("iterative_redundancy_pruning", "linear_svm", "mutual_information", "anova_f"),
    )
    cfgs = benchmark._build_ablation_configs(base, profile="core")
    names = [name for name, _ in cfgs]

    assert "fs_no_iterative_redundancy_pruning" in names


def test_iterative_pruning_bounded_ablation_toggle_present():
    base = benchmark.DFFSConfig(
        enabled_methods=("iterative_redundancy_pruning_bounded", "linear_svm", "mutual_information", "anova_f"),
    )
    cfgs = benchmark._build_ablation_configs(base, profile="core")
    names = [name for name, _ in cfgs]

    assert "fs_no_iterative_redundancy_pruning_bounded" in names
    assert "fs_enable_iterative_pruning_bounded_cpss_overlay" in names
    assert "fs_enable_iterative_pruning_class_pareto_prefilter" in names
    assert "fs_enable_iterative_pruning_class_pareto_stability_gate" in names


def test_tigress_ablation_toggle_present():
    base = benchmark.DFFSConfig(
        enabled_methods=("tigress_stability", "linear_svm", "mutual_information", "anova_f")
    )
    cfgs = benchmark._build_ablation_configs(base, profile="core")
    names = [name for name, _ in cfgs]

    assert "fs_no_tigress_stability" in names


def test_copula_stabilizer_ablation_toggle_present():
    base = benchmark.DFFSConfig(
        enabled_methods=("copula_knockoff", "linear_svm", "mutual_information", "anova_f"),
        fs_copula_stabilizer_runs=3,
        fs_copula_stabilizer_use_ebh=True,
    )
    cfgs = benchmark._build_ablation_configs(base, profile="core")
    names = [name for name, _ in cfgs]

    assert "fs_disable_copula_stabilizer" in names


def test_runtime_racing_ablation_toggle_present():
    base = benchmark.DFFSConfig(
        enabled_methods=("linear_svm", "mutual_information", "anova_f"),
        fs_runtime_racing_enabled=True,
    )
    cfgs = benchmark._build_ablation_configs(base, profile="core")
    names = [name for name, _ in cfgs]

    assert "fs_disable_runtime_racing" in names


def test_model_harness_flags_expand_candidates_in_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_easy_dfshift",
            "--include-elastic-net-model",
            "--include-rf-model",
            "--include-knn-model",
            "--include-xgb-model",
            "--include-tabpfn-model",
        ]
    )

    spec = benchmark.BENCHMARK_DATASETS["synthetic_easy_dfshift"]
    cfg = benchmark._build_base_config(args, spec, seed=11)

    assert cfg.model_candidates == ("lr", "svm_rbf", "elastic_net_lr", "rf", "knn", "xgb", "tabpfn")
    assert cfg.include_elastic_net_model is True
    assert cfg.include_rf_model is True
    assert cfg.include_knn_model is True
    assert cfg.include_xgb_model is True
    assert cfg.include_tabpfn_model is True


def test_benchmark_cli_defaults_are_strict():
    args = benchmark.build_arg_parser().parse_args(["--datasets", "synthetic_easy_dfshift"])
    assert args.allow_synthetic_fallback is False
    assert str(args.dataset_integrity_policy) == "error"


def test_benchmark_requires_hf_bundle_for_validation_catalog(monkeypatch, tmp_path):
    monkeypatch.delenv("TABNETICS_HF_ORG", raising=False)
    monkeypatch.delenv("TABNETICS_HF_REPO_ID", raising=False)

    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "cumida_brain_gse50161",
            "--seeds",
            "11",
            "--max-workers",
            "1",
            "--ablation-profile",
            "none",
            "--output-dir",
            str(tmp_path),
        ]
    )

    with pytest.raises(RuntimeError, match="HuggingFace bundle"):
        benchmark.run_benchmark(args)


def test_df_multimodal_defaults_are_enabled_in_benchmark_base_config():
    args = benchmark.build_arg_parser().parse_args(["--datasets", "synthetic_easy_dfshift"])
    spec = benchmark.BENCHMARK_DATASETS["synthetic_easy_dfshift"]
    cfg = benchmark._build_base_config(args, spec, seed=11)

    assert bool(args.df_compute_dip) is True
    assert str(args.df_multimodal_fallback) == "gmm"
    assert bool(cfg.dist_config.compute_dip) is True
    assert str(cfg.multimodal_fallback) == "gmm"


def test_df_multimodal_flags_allow_opt_out_and_rank_fallback():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_easy_dfshift",
            "--no-df-compute-dip",
            "--df-multimodal-fallback",
            "rank_transform",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_easy_dfshift"]
    cfg = benchmark._build_base_config(args, spec, seed=11)

    assert bool(args.df_compute_dip) is False
    assert bool(cfg.dist_config.compute_dip) is False
    assert str(cfg.multimodal_fallback) == "rank_transform"


def test_clone_config_preserves_df_multimodal_settings():
    base = benchmark.DFFSConfig(
        multimodal_fallback="rank_transform",
        dist_config=benchmark.DistributionFitterConfig(compute_dip=False),
    )
    cloned = benchmark.clone_config(base)
    assert bool(cloned.dist_config.compute_dip) is False
    assert str(cloned.multimodal_fallback) == "rank_transform"


@pytest.mark.skipif(
    not os.environ.get("TABNETICS_HF_ORG") or not os.environ.get("HF_TOKEN"),
    reason="TABNETICS_HF_ORG and HF_TOKEN must be set for integration tests",
)
def test_cumida_brain_succeeds_when_synthetic_fallback_disabled():
    """Integration test: verify that a real dataset loads successfully when synthetic fallback is disabled."""
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "cumida_brain_gse50161",
            "--seeds",
            "11",
            "--ablation-profile",
            "none",
        ]
    )

    result = benchmark._run_dataset_seed_task("cumida_brain_gse50161", 11, args)
    assert result["rows"] != []
    assert result["failures"] == []


def test_model_candidates_cli_override_controls_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_easy_dfshift",
            "--model-candidates",
            "xgb",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_easy_dfshift"]
    cfg = benchmark._build_base_config(args, spec, seed=11)

    assert cfg.model_candidates == ("xgb",)
    assert cfg.include_xgb_model is True


def test_model_candidate_profile_a11_controls_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_easy_dfshift",
            "--model-candidate-profile",
            "a11_medium_mismatch",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_easy_dfshift"]
    cfg = benchmark._build_base_config(args, spec, seed=11)

    assert cfg.model_candidates == ("lr", "svm_rbf", "svm_linear", "dlda", "knn", "nb", "vote_ensemble")
    assert cfg.include_svm_linear_model is True
    assert cfg.include_dlda_model is True
    assert cfg.include_knn_model is True
    assert cfg.include_nb_model is True
    assert cfg.include_vote_ensemble_model is True
    assert cfg.include_rf_model is False


def test_model_cv_runtime_containment_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_easy_dfshift",
            "--enable-model-cv-runtime-containment",
            "--model-cv-runtime-max-candidates",
            "3",
            "--model-cv-runtime-high-p-over-n-threshold",
            "25",
            "--model-cv-runtime-high-class-threshold",
            "5",
            "--model-cv-runtime-min-class-count-threshold",
            "10",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_easy_dfshift"]
    cfg = benchmark._build_base_config(args, spec, seed=11)

    assert cfg.model_cv_runtime_containment_enabled is True
    assert cfg.model_cv_runtime_max_candidates == 3
    assert cfg.model_cv_runtime_high_p_over_n_threshold == 25.0
    assert cfg.model_cv_runtime_high_class_threshold == 5
    assert cfg.model_cv_runtime_min_class_count_threshold == 10


def test_cumida_brain_benchmark_metadata_is_promotable():
    """CuMiDa brain is real-data; benchmark promotion metadata should not mark it fallback-only."""
    spec = benchmark.BENCHMARK_DATASETS["cumida_brain_gse50161"]
    meta = benchmark._benchmark_dataset_promotion_metadata(spec)

    assert int(meta["promotion_eligible"]) == 1
    assert str(meta["promotion_blocker"]) == ""
    assert str(meta["source_policy"]) in {"standard", "real_only"}


def test_wmw_auc_ablation_toggle_present():
    base = benchmark.DFFSConfig(
        enabled_methods=("wmw_auc", "linear_svm", "mutual_information", "anova_f"),
    )
    cfgs = benchmark._build_ablation_configs(base, profile="core")
    names = [name for name, _ in cfgs]

    assert "fs_no_wmw_auc" in names


def test_maqc_pairing_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_easy_dfshift",
            "--enable-maqc-pairing",
            "--maqc-fs-method-sets",
            "strict_plus_mrmr",
            "mnpo_ova_extended",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_easy_dfshift"]
    cfg = benchmark._build_base_config(args, spec, seed=11)

    assert cfg.enable_maqc_pairing is True
    assert cfg.maqc_pairing_method_set_names == ("strict_plus_mrmr", "mnpo_ova_extended")
    assert cfg.maqc_pairing_method_sets[0] == benchmark.FS_METHOD_SETS["strict_plus_mrmr"]
    assert cfg.maqc_pairing_method_sets[1] == benchmark.FS_METHOD_SETS["mnpo_ova_extended"]


def test_ova_linear_backend_flag_populates_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--fs-ova-linear-backend",
            "elastic_net_lr",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert cfg.fs_ova_linear_backend == "elastic_net_lr"


def test_ova_min_classes_flag_populates_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--fs-ova-min-classes",
            "5",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert int(cfg.fs_ova_min_classes) == 5


def test_ecoc_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--fs-ecoc-min-classes",
            "5",
            "--fs-ecoc-max-ovo-pairs",
            "9",
            "--fs-ecoc-random-code-bits",
            "3",
            "--fs-ecoc-class-complexity-weight",
            "1.3",
            "--fs-ecoc-negative-ratio",
            "1.6",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert int(cfg.fs_ecoc_min_classes) == 5
    assert int(cfg.fs_ecoc_max_ovo_pairs) == 9
    assert int(cfg.fs_ecoc_random_code_bits) == 3
    assert float(cfg.fs_ecoc_class_complexity_weight) == pytest.approx(1.3)
    assert float(cfg.fs_ecoc_negative_ratio) == pytest.approx(1.6)


def test_joint_multiclass_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--fs-joint-multiclass-min-classes",
            "4",
            "--fs-joint-multiclass-max-features",
            "180",
            "--fs-joint-multiclass-path-grid-size",
            "5",
            "--fs-joint-multiclass-min-c",
            "0.07",
            "--fs-joint-multiclass-max-c",
            "1.5",
            "--fs-joint-multiclass-l1-ratio",
            "0.6",
            "--fs-joint-multiclass-univariate-blend",
            "0.3",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert int(cfg.fs_joint_multiclass_min_classes) == 4
    assert int(cfg.fs_joint_multiclass_max_features) == 180
    assert int(cfg.fs_joint_multiclass_path_grid_size) == 5
    assert float(cfg.fs_joint_multiclass_min_c) == pytest.approx(0.07)
    assert float(cfg.fs_joint_multiclass_max_c) == pytest.approx(1.5)
    assert float(cfg.fs_joint_multiclass_l1_ratio) == pytest.approx(0.6)
    assert float(cfg.fs_joint_multiclass_univariate_blend) == pytest.approx(0.3)


def test_dove_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--fs-dove-min-classes",
            "4",
            "--fs-dove-max-pairs-per-class",
            "5",
            "--fs-dove-path-grid-size",
            "6",
            "--fs-dove-specificity-weight",
            "0.4",
            "--fs-dove-minority-boost",
            "0.7",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert int(cfg.fs_dove_min_classes) == 4
    assert int(cfg.fs_dove_max_pairs_per_class) == 5
    assert int(cfg.fs_dove_path_grid_size) == 6
    assert float(cfg.fs_dove_specificity_weight) == pytest.approx(0.4)
    assert float(cfg.fs_dove_minority_boost) == pytest.approx(0.7)


def test_sparse_multinomial_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--fs-sparse-multinomial-min-classes",
            "4",
            "--fs-sparse-multinomial-max-features",
            "180",
            "--fs-sparse-multinomial-path-grid-size",
            "5",
            "--fs-sparse-multinomial-min-c",
            "0.07",
            "--fs-sparse-multinomial-max-c",
            "1.4",
            "--fs-sparse-multinomial-backend",
            "elasticnet",
            "--fs-sparse-multinomial-l1-ratio",
            "0.6",
            "--fs-sparse-multinomial-univariate-blend",
            "0.3",
            "--fs-sparse-multinomial-max-iter",
            "4000",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert int(cfg.fs_sparse_multinomial_min_classes) == 4
    assert int(cfg.fs_sparse_multinomial_max_features) == 180
    assert int(cfg.fs_sparse_multinomial_path_grid_size) == 5
    assert float(cfg.fs_sparse_multinomial_min_c) == pytest.approx(0.07)
    assert float(cfg.fs_sparse_multinomial_max_c) == pytest.approx(1.4)
    assert str(cfg.fs_sparse_multinomial_backend) == "elasticnet"
    assert float(cfg.fs_sparse_multinomial_l1_ratio) == pytest.approx(0.6)
    assert float(cfg.fs_sparse_multinomial_univariate_blend) == pytest.approx(0.3)
    assert int(cfg.fs_sparse_multinomial_max_iter) == 4000


def test_nsc_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--fs-nsc-shrinkage-grid-size",
            "8",
            "--fs-nsc-min-classes",
            "4",
            "--fs-nsc-thresholding-mode",
            "auto",
            "--fs-nsc-order-quantile",
            "0.8",
            "--enable-fs-nsc-deep-shrinkage-search",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert int(cfg.fs_nsc_shrinkage_grid_size) == 8
    assert int(cfg.fs_nsc_min_classes) == 4
    assert str(cfg.fs_nsc_thresholding_mode) == "auto"
    assert float(cfg.fs_nsc_order_quantile) == pytest.approx(0.8)
    assert bool(cfg.fs_nsc_deep_shrinkage_search) is True


def test_class_pareto_and_hsic_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--fs-class-pareto-min-classes",
            "4",
            "--fs-class-pareto-top-per-class",
            "36",
            "--fs-class-pareto-global-fraction",
            "0.3",
            "--fs-class-pareto-minority-boost",
            "0.7",
            "--fs-class-pareto-kw-weight",
            "0.2",
            "--fs-hsic-lasso-alpha",
            "0.02",
            "--fs-hsic-lasso-prefilter-max-features",
            "96",
            "--fs-hsic-lasso-feature-sigma",
            "0.5",
            "--fs-hsic-lasso-target-sigma",
            "0.8",
            "--fs-hsic-lasso-relevance-blend",
            "0.3",
            "--fs-hsic-lasso-max-iter",
            "3000",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert int(cfg.fs_class_pareto_min_classes) == 4
    assert int(cfg.fs_class_pareto_top_per_class) == 36
    assert float(cfg.fs_class_pareto_global_fraction) == pytest.approx(0.3)
    assert float(cfg.fs_class_pareto_minority_boost) == pytest.approx(0.7)
    assert float(cfg.fs_class_pareto_kw_weight) == pytest.approx(0.2)
    assert float(cfg.fs_hsic_lasso_alpha) == pytest.approx(0.02)
    assert int(cfg.fs_hsic_lasso_prefilter_max_features) == 96
    assert float(cfg.fs_hsic_lasso_feature_sigma) == pytest.approx(0.5)
    assert float(cfg.fs_hsic_lasso_target_sigma) == pytest.approx(0.8)
    assert float(cfg.fs_hsic_lasso_relevance_blend) == pytest.approx(0.3)
    assert int(cfg.fs_hsic_lasso_max_iter) == 3000


def test_folding_and_face_projection_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--folding-method",
            "tensor_sketch",
            "--folding-n-components",
            "256",
            "--folding-rff-gamma",
            "0.25",
            "--folding-prefilter-k",
            "180",
            "--enable-face-domain-projection",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert str(cfg.folding_method) == "tensor_sketch"
    assert int(cfg.folding_n_components) == 256
    assert float(cfg.folding_rff_gamma) == pytest.approx(0.25)
    assert int(cfg.folding_prefilter_k) == 180
    assert bool(cfg.enable_face_domain_projection) is True


def test_runtime_racing_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--enable-fs-runtime-racing",
            "--fs-runtime-racing-proxy-splits",
            "2",
            "--fs-runtime-racing-keep-fraction",
            "0.5",
            "--fs-runtime-racing-min-candidates",
            "3",
            "--fs-runtime-racing-runtime-weight",
            "0.2",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert bool(cfg.fs_runtime_racing_enabled) is True
    assert int(cfg.fs_runtime_racing_proxy_splits) == 2
    assert float(cfg.fs_runtime_racing_keep_fraction) == pytest.approx(0.5)
    assert int(cfg.fs_runtime_racing_min_candidates) == 3
    assert float(cfg.fs_runtime_racing_runtime_weight) == pytest.approx(0.2)


def test_runtime_racing_op13_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--enable-fs-runtime-racing",
            "--fs-runtime-racing-mode",
            "successive_halving",
            "--fs-runtime-racing-stages",
            "3",
            "--fs-runtime-racing-confidence-bound",
            "bernstein",
            "--fs-runtime-racing-delta",
            "0.12",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert bool(cfg.fs_runtime_racing_enabled) is True
    assert str(cfg.fs_runtime_racing_mode) == "successive_halving"
    assert int(cfg.fs_runtime_racing_stages) == 3
    assert str(cfg.fs_runtime_racing_confidence_bound) == "bernstein"
    assert float(cfg.fs_runtime_racing_delta) == pytest.approx(0.12)


def test_pls_ova_sparse_quota_and_new_model_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--folding-method",
            "pls_da",
            "--folding-pls-components",
            "20",
            "--disable-folding-pls-scale",
            "--enable-fs-ova-calibration",
            "--fs-ova-calibration-cv",
            "4",
            "--fs-sparse-multinomial-screening-mode",
            "prefilter_aggressive",
            "--fs-sparse-multinomial-screening-keep-fraction",
            "0.6",
            "--fs-sparse-multinomial-screening-min-features",
            "72",
            "--disable-fs-sparse-multinomial-screening-fallback-on-failure",
            "--enable-fs-per-class-quota",
            "--fs-per-class-quota-min-per-class",
            "2",
            "--fs-per-class-quota-max-fraction",
            "0.5",
            "--include-nb-model",
            "--include-vote-ensemble-model",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert str(cfg.folding_method) == "pls_da"
    assert int(cfg.folding_pls_components) == 20
    assert bool(cfg.folding_pls_scale) is False
    assert bool(cfg.fs_ova_enable_calibration) is True
    assert int(cfg.fs_ova_calibration_cv) == 4
    assert str(cfg.fs_sparse_multinomial_screening_mode) == "prefilter_aggressive"
    assert float(cfg.fs_sparse_multinomial_screening_keep_fraction) == pytest.approx(0.6)
    assert int(cfg.fs_sparse_multinomial_screening_min_features) == 72
    assert bool(cfg.fs_sparse_multinomial_screening_fallback_on_failure) is False
    assert bool(cfg.fs_per_class_quota_enabled) is True
    assert int(cfg.fs_per_class_quota_min_per_class) == 2
    assert float(cfg.fs_per_class_quota_max_fraction) == pytest.approx(0.5)
    assert bool(cfg.include_nb_model) is True
    assert bool(cfg.include_vote_ensemble_model) is True


def test_tier_lockout_and_routing_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--tier-lockout-enabled",
            "--tier-lockout-tier",
            "easy",
            "--tier-lockout-difficulty-source",
            "historical",
            "--tier-lockout-fallback-fs-method-set",
            "strict_plus_mrmr",
            "--tier-routing-enabled",
            "--tier-routing-difficulty-classifier",
            "meta_features",
            "--tier-routing-table",
            "easy=strict_plus_mrmr;hard=linear_svm,anova_f",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert bool(cfg.tier_lockout_enabled) is True
    assert str(cfg.tier_lockout_tier) == "easy"
    assert str(cfg.tier_lockout_difficulty_source) == "historical"
    assert tuple(cfg.tier_lockout_fallback_methods) == tuple(benchmark.FS_METHOD_SETS["strict_plus_mrmr"])
    assert bool(cfg.tier_routing_enabled) is True
    assert str(cfg.tier_routing_difficulty_classifier) == "meta_features"
    assert tuple(cfg.tier_routing_table["easy"]) == tuple(benchmark.FS_METHOD_SETS["strict_plus_mrmr"])
    assert tuple(cfg.tier_routing_table["hard"]) == ("linear_svm", "anova_f")


def test_regime_gating_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--regime-gating-enabled",
            "--regime-gating-difficulty-source",
            "historical",
            "--regime-gating-target-tier",
            "very_hard",
            "--regime-gating-min-samples-per-class",
            "15",
            "--regime-gating-low-p-over-n-threshold",
            "2.0",
            "--regime-gating-simple-fs-method-set",
            "strict_plus_mrmr",
            "--regime-gating-very-hard-portfolio-max-methods",
            "4",
            "--regime-gating-very-hard-copula-derandomize-runs",
            "5",
            "--regime-gating-low-p-over-n-mode",
            "all_features",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert bool(cfg.regime_gating_enabled) is True
    assert str(cfg.regime_gating_difficulty_source) == "historical"
    assert str(cfg.regime_gating_target_tier) == "very_hard"
    assert float(cfg.regime_gating_min_samples_per_class) == pytest.approx(15.0)
    assert float(cfg.regime_gating_low_p_over_n_threshold) == pytest.approx(2.0)
    assert tuple(cfg.regime_gating_simple_methods) == tuple(benchmark.FS_METHOD_SETS["strict_plus_mrmr"])
    assert int(cfg.regime_gating_very_hard_portfolio_max_methods) == 4
    assert int(cfg.regime_gating_very_hard_copula_derandomize_runs) == 5
    assert str(cfg.regime_gating_low_p_over_n_mode) == "all_features"


def test_importance_uq_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--fs-importance-uq-enabled",
            "--fs-importance-uq-min-cv-folds",
            "5",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert bool(cfg.fs_importance_uq_enabled) is True
    assert int(cfg.fs_importance_uq_min_cv_folds) == 5


def test_copula_deepdrk_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--fs-copula-generator",
            "deepdrk",
            "--fs-copula-deepdrk-latent-fraction",
            "0.4",
            "--fs-copula-deepdrk-noise-scale",
            "1.2",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert str(cfg.fs_copula_generator) == "deepdrk"
    assert float(cfg.fs_copula_deepdrk_latent_fraction) == pytest.approx(0.4)
    assert float(cfg.fs_copula_deepdrk_noise_scale) == pytest.approx(1.2)


def test_adaptive_portfolio_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--fs-portfolio-size",
            "6",
            "--enable-fs-adaptive-portfolio-sizing",
            "--fs-adaptive-size-min",
            "4",
            "--fs-adaptive-size-max",
            "8",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert int(cfg.fs_portfolio_size) == 6
    assert bool(cfg.fs_adaptive_portfolio_sizing_enabled) is True
    assert int(cfg.fs_adaptive_size_min) == 4
    assert int(cfg.fs_adaptive_size_max) == 8


def test_val7_shapley_and_adaptive_penalty_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--enable-fs-adaptive-portfolio-sizing",
            "--adaptive-sizing-variance-penalty",
            "--adaptive-sizing-variance-penalty-strength",
            "0.7",
            "--fs-oracle-weighting-mode",
            "shapley",
            "--shapley-bayesian-shrinkage",
            "--shapley-bayesian-prior-strength",
            "9.5",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert bool(cfg.fs_adaptive_portfolio_sizing_enabled) is True
    assert bool(cfg.fs_adaptive_sizing_variance_penalty) is True
    assert float(cfg.fs_adaptive_sizing_variance_penalty_strength) == pytest.approx(0.7)
    assert str(cfg.fs_oracle_weighting_mode) == "shapley"
    assert bool(cfg.fs_shapley_bayesian_shrinkage) is True
    assert float(cfg.fs_shapley_bayesian_prior_strength) == pytest.approx(9.5)


def test_fs_portfolio_size_default_is_six():
    args = benchmark.build_arg_parser().parse_args(["--datasets", "synthetic_medium_mixed"])
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert int(cfg.fs_portfolio_size) == 6


def test_adaptive_portfolio_defaults_to_plus_minus_two_when_enabled_without_bounds():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--fs-portfolio-size",
            "7",
            "--enable-fs-adaptive-portfolio-sizing",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert bool(cfg.fs_adaptive_portfolio_sizing_enabled) is True
    assert int(cfg.fs_adaptive_size_min) == 5
    assert int(cfg.fs_adaptive_size_max) == 9


def test_rashomon_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--enable-fs-rashomon",
            "--fs-rashomon-max-models",
            "14",
            "--fs-rashomon-score-tolerance",
            "0.02",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert bool(cfg.fs_rashomon_enabled) is True
    assert int(cfg.fs_rashomon_max_models) == 14
    assert float(cfg.fs_rashomon_score_tolerance) == pytest.approx(0.02)


def test_iterative_pruning_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--fs-iterative-pruning-pool-factor",
            "2.2",
            "--fs-iterative-pruning-max-rounds",
            "15",
            "--fs-iterative-pruning-min-improvement",
            "-0.01",
            "--fs-iterative-pruning-max-cumulative-loss",
            "0.015",
            "--fs-iterative-pruning-redundancy-weight",
            "0.7",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert float(cfg.fs_iterative_pruning_pool_factor) == pytest.approx(2.2)
    assert int(cfg.fs_iterative_pruning_max_rounds) == 15
    assert float(cfg.fs_iterative_pruning_min_improvement) == pytest.approx(-0.01)
    assert float(cfg.fs_iterative_pruning_max_cumulative_loss) == pytest.approx(0.015)
    assert float(cfg.fs_iterative_pruning_redundancy_weight) == pytest.approx(0.7)


def test_iterative_pruning_bounded_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--fs-iterative-pruning-bounded-prefilter-cap",
            "180",
            "--fs-iterative-pruning-bounded-candidate-fraction",
            "0.28",
            "--fs-iterative-pruning-bounded-min-candidates",
            "3",
            "--fs-iterative-pruning-bounded-max-evaluations",
            "16",
            "--fs-iterative-pruning-bounded-max-runtime-seconds",
            "12.0",
            "--fs-iterative-pruning-bounded-multiclass-scale",
            "0.6",
            "--fs-iterative-pruning-bounded-imbalance-trigger",
            "2.1",
            "--fs-iterative-pruning-bounded-imbalance-scale",
            "0.68",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert int(cfg.fs_iterative_pruning_bounded_prefilter_cap) == 180
    assert float(cfg.fs_iterative_pruning_bounded_candidate_fraction) == pytest.approx(0.28)
    assert int(cfg.fs_iterative_pruning_bounded_min_candidates) == 3
    assert int(cfg.fs_iterative_pruning_bounded_max_evaluations) == 16
    assert float(cfg.fs_iterative_pruning_bounded_max_runtime_seconds) == pytest.approx(12.0)
    assert float(cfg.fs_iterative_pruning_bounded_multiclass_scale) == pytest.approx(0.6)
    assert float(cfg.fs_iterative_pruning_bounded_imbalance_trigger) == pytest.approx(2.1)
    assert float(cfg.fs_iterative_pruning_bounded_imbalance_scale) == pytest.approx(0.68)
    assert bool(cfg.fs_iterative_pruning_bounded_enable_class_gating) is True


def test_iterative_pruning_bounded_cpss_and_pareto_flags_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--enable-fs-iterative-pruning-bounded-cpss-overlay",
            "--fs-iterative-pruning-bounded-cpss-pairs",
            "3",
            "--fs-iterative-pruning-bounded-cpss-stability-threshold",
            "0.58",
            "--fs-iterative-pruning-bounded-cpss-min-stable-features",
            "2",
            "--fs-iterative-pruning-bounded-cpss-min-jaccard",
            "0.22",
            "--fs-iterative-pruning-bounded-cpss-max-score-drop",
            "0.015",
            "--enable-fs-iterative-pruning-class-pareto-prefilter",
            "--fs-iterative-pruning-class-pareto-min-classes",
            "3",
            "--fs-iterative-pruning-class-pareto-top-per-class",
            "20",
            "--fs-iterative-pruning-class-pareto-global-fraction",
            "0.33",
            "--fs-iterative-pruning-class-pareto-minority-boost",
            "0.7",
            "--enable-fs-iterative-pruning-class-pareto-stability-gate",
            "--fs-iterative-pruning-class-pareto-stability-subsamples",
            "5",
            "--fs-iterative-pruning-class-pareto-stability-fraction",
            "0.72",
            "--fs-iterative-pruning-class-pareto-stability-threshold",
            "0.56",
            "--fs-iterative-pruning-class-pareto-stability-min-overlap",
            "0.45",
            "--fs-iterative-pruning-class-pareto-stability-min-stable-features",
            "2",
            "--disable-fs-iterative-pruning-class-pareto-stability-fallback-on-failure",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert bool(cfg.fs_iterative_pruning_bounded_use_cpss_overlay) is True
    assert int(cfg.fs_iterative_pruning_bounded_cpss_pairs) == 3
    assert float(cfg.fs_iterative_pruning_bounded_cpss_stability_threshold) == pytest.approx(0.58)
    assert float(cfg.fs_iterative_pruning_bounded_cpss_min_jaccard) == pytest.approx(0.22)
    assert float(cfg.fs_iterative_pruning_bounded_cpss_max_score_drop) == pytest.approx(0.015)
    assert bool(cfg.fs_iterative_pruning_class_pareto_prefilter_enabled) is True
    assert int(cfg.fs_iterative_pruning_class_pareto_min_classes) == 3
    assert int(cfg.fs_iterative_pruning_class_pareto_top_per_class) == 20
    assert float(cfg.fs_iterative_pruning_class_pareto_global_fraction) == pytest.approx(0.33)
    assert float(cfg.fs_iterative_pruning_class_pareto_minority_boost) == pytest.approx(0.7)
    assert bool(cfg.fs_iterative_pruning_class_pareto_stability_gate_enabled) is True
    assert int(cfg.fs_iterative_pruning_class_pareto_stability_subsamples) == 5
    assert float(cfg.fs_iterative_pruning_class_pareto_stability_fraction) == pytest.approx(0.72)
    assert float(cfg.fs_iterative_pruning_class_pareto_stability_threshold) == pytest.approx(0.56)
    assert float(cfg.fs_iterative_pruning_class_pareto_stability_min_overlap) == pytest.approx(0.45)
    assert int(cfg.fs_iterative_pruning_class_pareto_stability_min_stable_features) == 2
    assert bool(cfg.fs_iterative_pruning_class_pareto_stability_fallback_on_failure) is False


def test_iterative_pruning_variant_method_sets_enable_expected_overlays_by_default():
    parser = benchmark.build_arg_parser()
    spec = benchmark.BENCHMARK_DATASETS["synthetic_medium_mixed"]

    args_base = parser.parse_args(
        ["--datasets", "synthetic_medium_mixed", "--fs-method-set", "mnpo_iterative_pruning_bounded_extended"]
    )
    cfg_base = benchmark._build_base_config(args_base, spec, seed=11)
    assert bool(cfg_base.fs_iterative_pruning_bounded_use_cpss_overlay) is False
    assert bool(cfg_base.fs_iterative_pruning_class_pareto_prefilter_enabled) is False
    assert bool(cfg_base.fs_iterative_pruning_class_pareto_stability_gate_enabled) is False

    args_cpss = parser.parse_args(
        ["--datasets", "synthetic_medium_mixed", "--fs-method-set", "mnpo_iterative_pruning_bounded_cpss_extended"]
    )
    cfg_cpss = benchmark._build_base_config(args_cpss, spec, seed=11)
    assert bool(cfg_cpss.fs_iterative_pruning_bounded_use_cpss_overlay) is True
    assert bool(cfg_cpss.fs_iterative_pruning_class_pareto_prefilter_enabled) is False
    assert bool(cfg_cpss.fs_iterative_pruning_class_pareto_stability_gate_enabled) is False

    args_pareto = parser.parse_args(
        ["--datasets", "synthetic_medium_mixed", "--fs-method-set", "mnpo_iterative_pruning_bounded_pareto_extended"]
    )
    cfg_pareto = benchmark._build_base_config(args_pareto, spec, seed=11)
    assert bool(cfg_pareto.fs_iterative_pruning_bounded_use_cpss_overlay) is False
    assert bool(cfg_pareto.fs_iterative_pruning_class_pareto_prefilter_enabled) is True
    assert bool(cfg_pareto.fs_iterative_pruning_class_pareto_stability_gate_enabled) is False

    args_stability = parser.parse_args(
        [
            "--datasets",
            "synthetic_medium_mixed",
            "--fs-method-set",
            "mnpo_iterative_pruning_bounded_pareto_stability_extended",
        ]
    )
    cfg_stability = benchmark._build_base_config(args_stability, spec, seed=11)
    assert bool(cfg_stability.fs_iterative_pruning_bounded_use_cpss_overlay) is False
    assert bool(cfg_stability.fs_iterative_pruning_class_pareto_prefilter_enabled) is True
    assert bool(cfg_stability.fs_iterative_pruning_class_pareto_stability_gate_enabled) is True

def test_df_fastpath_flags_are_noop_after_cleanup():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "int_low_gof_downweighting",
            "--enable-df-fastpath",
            "--df-fastpath-scope",
            "fs_only",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["int_low_gof_downweighting"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert cfg.df_fastpath_enabled is False

    args_fs = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "lung_gordon",
            "--enable-df-fastpath",
            "--df-fastpath-scope",
            "fs_only",
        ]
    )
    spec_fs = benchmark.BENCHMARK_DATASETS["lung_gordon"]
    cfg_fs = benchmark._build_base_config(args_fs, spec_fs, seed=11)
    assert cfg_fs.df_fastpath_enabled is False


def test_df_fastpath_cli_flags_emit_deprecation_warnings():
    benchmark._DEPRECATED_TOGGLE_WARNED.clear()
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_easy_dfshift",
            "--enable-df-fastpath",
            "--df-fastpath-trigger",
            "low_unique",
            "--df-fastpath-small-n-threshold",
            "128",
            "--df-fastpath-unique-ratio-threshold",
            "0.10",
            "--df-fastpath-n-unique-threshold",
            "3",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_easy_dfshift"]
    with pytest.warns(DeprecationWarning):
        _ = benchmark._build_base_config(args, spec, seed=11)


def test_maqc_pairing_improvement_thresholds_populate_base_config():
    args = benchmark.build_arg_parser().parse_args(
        [
            "--datasets",
            "synthetic_easy_dfshift",
            "--enable-maqc-pairing",
            "--maqc-fs-method-sets",
            "strict_plus_mrmr",
            "mnpo_ova_extended",
            "--maqc-pairing-min-improvement",
            "0.02",
            "--maqc-pairing-min-improvement-se-mult",
            "1.0",
        ]
    )
    spec = benchmark.BENCHMARK_DATASETS["synthetic_easy_dfshift"]
    cfg = benchmark._build_base_config(args, spec, seed=11)
    assert cfg.enable_maqc_pairing is True
    assert float(cfg.maqc_pairing_min_improvement) == pytest.approx(0.02)
    assert float(cfg.maqc_pairing_min_improvement_se_mult) == pytest.approx(1.0)



"""Tests for multi-omics integration (VAL12_Suggestions §4.1)."""

import numpy as np
import pytest

from tabnetics.multiomics.integration import MultiBlockPLSDA, MINTIntegrator


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def _make_two_block_data(n=120, p1=10, p2=8, seed=42):
    """Create synthetic 2-block data with class-discriminating signal."""
    rng = np.random.RandomState(seed)
    y = np.tile([0, 1, 2], int(np.ceil(n / 3)))[:n]

    # Block 1: features 0-2 discriminate class 0.
    X1 = rng.randn(n, p1)
    X1[y == 0, :3] += 2.0

    # Block 2: features 0-2 discriminate class 1.
    X2 = rng.randn(n, p2)
    X2[y == 1, :3] += 2.0

    return X1, X2, y


# ---------------------------------------------------------------------------
# MultiBlockPLSDA tests
# ---------------------------------------------------------------------------


class TestMultiBlockPLSDA:
    """Verify multi-block PLS-DA fitting and transformation."""

    def test_fit_returns_self(self):
        X1, X2, y = _make_two_block_data()
        model = MultiBlockPLSDA(n_components=2)
        result = model.fit([(X1, "block1"), (X2, "block2")], y)
        assert result is model

    def test_super_scores_shape(self):
        X1, X2, y = _make_two_block_data(n=60)
        model = MultiBlockPLSDA(n_components=3)
        model.fit([(X1, "block1"), (X2, "block2")], y)
        assert model.super_scores_.shape == (60, 3)

    def test_block_loadings_shapes(self):
        X1, X2, y = _make_two_block_data()
        model = MultiBlockPLSDA(n_components=2)
        model.fit([(X1, "b1"), (X2, "b2")], y)
        assert len(model.block_loadings_) == 2
        assert model.block_loadings_[0].shape == (X1.shape[1], 2)
        assert model.block_loadings_[1].shape == (X2.shape[1], 2)

    def test_transform_shape(self):
        X1, X2, y = _make_two_block_data(n=120)
        model = MultiBlockPLSDA(n_components=2)
        model.fit([(X1[:80], "b1"), (X2[:80], "b2")], y[:80])
        T = model.transform([(X1[80:], "b1"), (X2[80:], "b2")])
        assert T.shape == (40, 2)

    def test_fit_transform_equals_super_scores(self):
        X1, X2, y = _make_two_block_data()
        model = MultiBlockPLSDA(n_components=2)
        T = model.fit_transform([(X1, "b1"), (X2, "b2")], y)
        np.testing.assert_array_equal(T, model.super_scores_)

    def test_feature_importance_per_block(self):
        X1, X2, y = _make_two_block_data()
        model = MultiBlockPLSDA(n_components=2)
        model.fit([(X1, "block_a"), (X2, "block_b")], y)
        imp = model.get_feature_importance()
        assert "block_a" in imp
        assert "block_b" in imp
        assert imp["block_a"].shape == (X1.shape[1],)
        assert imp["block_b"].shape == (X2.shape[1],)

    def test_informative_features_have_higher_importance(self):
        """Discriminating features should have higher importance."""
        X1, X2, y = _make_two_block_data(n=200, seed=7)
        model = MultiBlockPLSDA(n_components=2)
        model.fit([(X1, "b1"), (X2, "b2")], y)
        imp = model.get_feature_importance()
        # Block 1: features 0-2 are informative.
        info_imp = np.mean(imp["b1"][:3])
        noise_imp = np.mean(imp["b1"][3:])
        assert info_imp > noise_imp, (
            f"Informative features ({info_imp:.4f}) should have higher "
            f"importance than noise ({noise_imp:.4f})"
        )

    def test_latent_scores_improve_separability(self):
        """Latent scores should yield better class separation than random."""
        X1, X2, y = _make_two_block_data(n=120, seed=13)
        model = MultiBlockPLSDA(n_components=2)
        T = model.fit_transform([(X1, "b1"), (X2, "b2")], y)
        # Compute between-class variance / within-class variance.
        classes = np.unique(y)
        grand_mean = np.mean(T[:, 0])
        between = sum(
            np.sum(y == c) * (np.mean(T[y == c, 0]) - grand_mean) ** 2
            for c in classes
        )
        within = sum(np.sum((T[y == c, 0] - np.mean(T[y == c, 0])) ** 2) for c in classes)
        ratio = between / max(1e-12, within)
        # Should have some class separation.
        assert ratio > 0.01, f"Fisher ratio too low: {ratio:.4f}"

    def test_single_block_works(self):
        X1, _, y = _make_two_block_data(n=60)
        model = MultiBlockPLSDA(n_components=2)
        model.fit([(X1, "only")], y)
        assert model.super_scores_.shape == (60, 2)

    def test_binary_classification(self):
        rng = np.random.RandomState(42)
        n = 80
        X1 = rng.randn(n, 6)
        X2 = rng.randn(n, 4)
        y = np.repeat([0, 1], 40)
        X1[:40, :2] += 1.5
        X2[:40, :2] -= 1.5
        model = MultiBlockPLSDA(n_components=2)
        model.fit([(X1, "b1"), (X2, "b2")], y)
        assert model.super_scores_.shape == (80, 2)

    def test_deterministic(self):
        X1, X2, y = _make_two_block_data()
        m1 = MultiBlockPLSDA(n_components=2)
        m1.fit([(X1, "b1"), (X2, "b2")], y)
        m2 = MultiBlockPLSDA(n_components=2)
        m2.fit([(X1, "b1"), (X2, "b2")], y)
        np.testing.assert_allclose(m1.super_scores_, m2.super_scores_, atol=1e-10)

    def test_explained_variance_nonnegative(self):
        X1, X2, y = _make_two_block_data()
        model = MultiBlockPLSDA(n_components=2)
        model.fit([(X1, "b1"), (X2, "b2")], y)
        assert np.all(model.explained_variance_ >= 0.0)

    def test_mismatched_samples_raises(self):
        X1, X2, y = _make_two_block_data(n=60)
        with pytest.raises(ValueError, match="samples"):
            MultiBlockPLSDA().fit([(X1[:50], "b1"), (X2, "b2")], y)

    def test_mismatched_y_raises(self):
        X1, X2, y = _make_two_block_data(n=60)
        with pytest.raises(ValueError, match="y length"):
            MultiBlockPLSDA().fit([(X1, "b1"), (X2, "b2")], y[:30])

    def test_unfitted_transform_raises(self):
        model = MultiBlockPLSDA()
        with pytest.raises(RuntimeError, match="not fitted"):
            model.transform([(np.zeros((5, 3)), "b1")])


# ---------------------------------------------------------------------------
# MINTIntegrator tests
# ---------------------------------------------------------------------------


class TestMINTIntegrator:
    """Verify MINT-style study-aware integration."""

    def test_fit_returns_self(self):
        X1, X2, y = _make_two_block_data(n=120)
        studies = np.array(["A"] * 60 + ["B"] * 60)
        model = MINTIntegrator(n_components=2)
        result = model.fit([(X1, "b1"), (X2, "b2")], y, studies)
        assert result is model

    def test_transform_shape(self):
        X1, X2, y = _make_two_block_data(n=120)
        studies = np.array(["A"] * 60 + ["B"] * 60)
        model = MINTIntegrator(n_components=2)
        model.fit([(X1[:80], "b1"), (X2[:80], "b2")], y[:80], studies[:80])
        T = model.transform([(X1[80:], "b1"), (X2[80:], "b2")], studies[80:])
        assert T.shape == (40, 2)

    def test_study_centering_reduces_batch_effect(self):
        """MINT should reduce study-induced mean shifts."""
        rng = np.random.RandomState(42)
        n = 120
        y = np.repeat([0, 1], 60)
        X1 = rng.randn(n, 6)
        X2 = rng.randn(n, 4)
        X1[:60, :2] += 1.0  # class signal
        studies = np.array(["A"] * 60 + ["B"] * 60)
        # Inject study effect.
        X1[studies == "B"] += 3.0
        X2[studies == "B"] += 3.0

        model_no_mint = MultiBlockPLSDA(n_components=2)
        T_no = model_no_mint.fit_transform([(X1, "b1"), (X2, "b2")], y)

        model_mint = MINTIntegrator(n_components=2)
        T_mint = model_mint.fit_transform(
            [(X1, "b1"), (X2, "b2")], y, studies,
        )

        # Study-mean gap in latent space should be smaller with MINT.
        gap_no = abs(np.mean(T_no[studies == "A", 0]) - np.mean(T_no[studies == "B", 0]))
        gap_mint = abs(np.mean(T_mint[studies == "A", 0]) - np.mean(T_mint[studies == "B", 0]))
        assert gap_mint <= gap_no + 0.1, (
            f"MINT should reduce study gap: {gap_mint:.3f} > {gap_no:.3f}"
        )

    def test_feature_importance_accessible(self):
        X1, X2, y = _make_two_block_data(n=120)
        studies = np.array(["A"] * 60 + ["B"] * 60)
        model = MINTIntegrator(n_components=2)
        model.fit([(X1, "b1"), (X2, "b2")], y, studies)
        imp = model.get_feature_importance()
        assert "b1" in imp
        assert "b2" in imp

    def test_unseen_study_handled(self):
        """Transform with unseen study should not crash."""
        X1, X2, y = _make_two_block_data(n=120)
        studies_train = np.array(["A"] * 60 + ["B"] * 60)
        model = MINTIntegrator(n_components=2)
        model.fit([(X1, "b1"), (X2, "b2")], y, studies_train)
        studies_test = np.array(["C"] * 40)  # unseen
        T = model.transform([(X1[80:], "b1"), (X2[80:], "b2")], studies_test)
        assert T.shape == (40, 2)
        assert np.all(np.isfinite(T))

    def test_deterministic(self):
        X1, X2, y = _make_two_block_data(n=60)
        studies = np.array(["A"] * 30 + ["B"] * 30)
        m1 = MINTIntegrator(n_components=2)
        T1 = m1.fit_transform([(X1, "b1"), (X2, "b2")], y, studies)
        m2 = MINTIntegrator(n_components=2)
        T2 = m2.fit_transform([(X1, "b1"), (X2, "b2")], y, studies)
        np.testing.assert_allclose(T1, T2, atol=1e-10)

    def test_unfitted_raises(self):
        model = MINTIntegrator()
        with pytest.raises(RuntimeError, match="not fitted"):
            model.transform(
                [(np.zeros((5, 3)), "b1")],
                np.array(["A"] * 5),
            )

    def test_study_labels_mismatch_raises(self):
        X1, X2, y = _make_two_block_data(n=60)
        with pytest.raises(ValueError, match="mismatch"):
            MINTIntegrator().fit(
                [(X1, "b1"), (X2, "b2")], y,
                np.array(["A"] * 30),  # wrong length
            )



"""Tests for T-R-249: Banzhaf value computation."""

import unittest
import numpy as np


class TestBanzhafValues(unittest.TestCase):
    """Unit tests for Banzhaf value computation in mnpo_core."""

    def _make_oracle_matrices(self, k=3, m=4):
        """Create k random oracle preference matrices of size m×m."""
        rng = np.random.RandomState(42)
        matrices = {}
        for i in range(k):
            mat = rng.uniform(0.1, 0.9, (m, m))
            np.fill_diagonal(mat, 0.5)
            matrices[f"oracle_{i}"] = mat
        return matrices

    def test_empty_oracles_returns_empty(self):
        from tabnetics.core.mnpo import compute_banzhaf_values
        weights, meta = compute_banzhaf_values({})
        self.assertEqual(weights, {})
        self.assertFalse(meta["applied"])

    def test_single_oracle_returns_unit_weight(self):
        from tabnetics.core.mnpo import compute_banzhaf_values
        mat = np.full((3, 3), 0.5)
        weights, meta = compute_banzhaf_values({"perf": mat})
        self.assertEqual(weights["perf"], 1.0)
        self.assertTrue(meta["applied"])

    def test_banzhaf_weights_sum_to_one(self):
        from tabnetics.core.mnpo import compute_banzhaf_values
        matrices = self._make_oracle_matrices(k=3, m=4)
        weights, meta = compute_banzhaf_values(matrices)
        self.assertTrue(meta["applied"])
        self.assertAlmostEqual(sum(weights.values()), 1.0, places=8)

    def test_banzhaf_values_are_non_negative(self):
        from tabnetics.core.mnpo import compute_banzhaf_values
        matrices = self._make_oracle_matrices(k=4, m=5)
        weights, meta = compute_banzhaf_values(matrices)
        for w in weights.values():
            self.assertGreaterEqual(w, 0.0)

    def test_banzhaf_raw_values_in_meta(self):
        from tabnetics.core.mnpo import compute_banzhaf_values
        matrices = self._make_oracle_matrices(k=3, m=4)
        _, meta = compute_banzhaf_values(matrices)
        self.assertIn("raw_banzhaf", meta)
        self.assertEqual(len(meta["raw_banzhaf"]), 3)

    def test_kernel_banzhaf_fallback_for_many_oracles(self):
        from tabnetics.core.mnpo import compute_banzhaf_values
        matrices = self._make_oracle_matrices(k=14, m=4)
        weights, meta = compute_banzhaf_values(matrices, max_coalitions=256)
        self.assertTrue(meta["applied"])
        self.assertEqual(meta["method"], "kernel_regression")
        self.assertAlmostEqual(sum(weights.values()), 1.0, places=6)

    def test_kernel_banzhaf_direct(self):
        from tabnetics.core.mnpo import kernel_banzhaf_values
        matrices = self._make_oracle_matrices(k=5, m=4)
        weights, meta = kernel_banzhaf_values(matrices, n_samples=512, seed=0)
        self.assertTrue(meta["applied"])
        self.assertAlmostEqual(sum(weights.values()), 1.0, places=6)
        for w in weights.values():
            self.assertGreaterEqual(w, 0.0)

    def test_banzhaf_exact_vs_kernel_agreement(self):
        """Exact and kernel Banzhaf should roughly agree for small k."""
        from tabnetics.core.mnpo import compute_banzhaf_values, kernel_banzhaf_values
        matrices = self._make_oracle_matrices(k=3, m=6)
        exact_w, _ = compute_banzhaf_values(matrices)
        kernel_w, _ = kernel_banzhaf_values(matrices, n_samples=4096, seed=42)
        # Kernel estimator has sampling variance; 0.15 delta is expected
        # for 4096 samples with k=3 oracles.
        for name in exact_w:
            self.assertAlmostEqual(exact_w[name], kernel_w[name], delta=0.15)

    def test_identical_oracles_get_equal_banzhaf(self):
        """Two identical oracles should get equal Banzhaf values."""
        from tabnetics.core.mnpo import compute_banzhaf_values
        mat = np.random.RandomState(7).uniform(0.1, 0.9, (4, 4))
        np.fill_diagonal(mat, 0.5)
        matrices = {"a": mat.copy(), "b": mat.copy(), "ref": np.full((4, 4), 0.5)}
        weights, _ = compute_banzhaf_values(matrices, reference="ref")
        self.assertAlmostEqual(weights["a"], weights["b"], places=8)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import time

import numpy as np
import pandas as pd
import pytest

import tabnetics.distribution.selector as uds_mod
from tabnetics.feature_selection.base import FeatureSelector
from tabnetics.feature_selection.methods.embedded import linear_svm_selection
from tabnetics.feature_selection.methods.filter import (
    anova_f_selection,
    chi_square_selection,
    mrmr_jmi_selection,
    mutual_information_selection,
    relieff_selection,
    wmw_auc_selection,
)
from tabnetics.feature_selection.methods.knockoff import copula_knockoff_selection
from tabnetics.feature_selection.methods.pairwise import ktsp_selection
from tabnetics.feature_selection.preprocessing import remove_correlated_features


def _toy_xy(n: int, p: int, *, seed: int = 0) -> tuple[np.ndarray, np.ndarray]:
    rng = np.random.default_rng(seed)
    X = rng.normal(size=(int(n), int(p)))
    if int(n) <= 0:
        y = np.asarray([], dtype=int)
    else:
        y = np.asarray([i % 2 for i in range(int(n))], dtype=int)
    return X, y


def _var_scorer(X: np.ndarray, y: np.ndarray, random_state: int | None = None) -> np.ndarray:
    del y, random_state
    if X.size == 0:
        return np.asarray([], dtype=float)
    return np.nan_to_num(np.var(X, axis=0), nan=0.0, posinf=0.0, neginf=0.0)


def _f_scorer(X: np.ndarray, y: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    del y
    scores = _var_scorer(X, np.asarray([]))
    return scores, np.zeros(scores.shape[0], dtype=float)


def _normalize(scores: np.ndarray) -> np.ndarray:
    arr = np.asarray(scores, dtype=float).ravel()
    if arr.size == 0:
        return arr
    lo = float(np.min(arr))
    hi = float(np.max(arr))
    if not np.isfinite(lo) or not np.isfinite(hi) or hi <= lo:
        return np.zeros_like(arr, dtype=float)
    return (arr - lo) / (hi - lo)


def _prefilter_small(X: np.ndarray, y: np.ndarray, max_features: int) -> np.ndarray:
    del y
    n_features = int(np.asarray(X).shape[1]) if np.asarray(X).ndim == 2 else 0
    keep = int(min(n_features, max(0, min(int(max_features), 16))))
    return np.arange(keep, dtype=int)


def _call_mutual_information(X: np.ndarray, y: np.ndarray):
    return mutual_information_selection(X, y, 5, _var_scorer, 0)


def _call_anova(X: np.ndarray, y: np.ndarray):
    return anova_f_selection(X, y, 5, _f_scorer)


def _call_wmw_auc(X: np.ndarray, y: np.ndarray):
    return wmw_auc_selection(X, y, 5, "classification")


def _call_chi_square(X: np.ndarray, y: np.ndarray):
    return chi_square_selection(X, y, 5)


def _call_relieff(X: np.ndarray, y: np.ndarray):
    return relieff_selection(X, y, 5, n_neighbors=3)


def _call_mrmr(X: np.ndarray, y: np.ndarray):
    return mrmr_jmi_selection(
        X,
        y,
        5,
        random_state=0,
        mi_scorer=_var_scorer,
        prefilter_fn=_prefilter_small,
        normalize_fn=_normalize,
        mrmr_max_features=16,
        mrmr_redundancy_weight=0.5,
    )


def _call_ktsp(X: np.ndarray, y: np.ndarray):
    return ktsp_selection(
        X,
        y,
        5,
        problem_type="classification",
        random_state=0,
        prefilter_fn=_prefilter_small,
        ktsp_max_features=16,
        ktsp_max_pairs=64,
        ktsp_k_pairs=8,
    )


def _call_linear_svm(X: np.ndarray, y: np.ndarray):
    return linear_svm_selection(
        X,
        y,
        5,
        n_bootstrap_iterations=1,
        problem_type="classification",
        random_state=0,
        linear_svm_max_iter=300,
    )


METHOD_CALLS = {
    "mutual_information": _call_mutual_information,
    "anova_f": _call_anova,
    "wmw_auc": _call_wmw_auc,
    "chi_square": _call_chi_square,
    "relieff": _call_relieff,
    "mrmr_jmi": _call_mrmr,
    "ktsp": _call_ktsp,
    "linear_svm": _call_linear_svm,
}


@pytest.mark.parametrize("method_name", sorted(METHOD_CALLS.keys()))
@pytest.mark.parametrize("n", [0, 1, 2, 5])
@pytest.mark.parametrize("p", [0, 1, 500])
def test_parametrized_edge_cases_method_entry_points(method_name: str, n: int, p: int) -> None:
    X, y = _toy_xy(n, p, seed=(n * 1000 + p + len(method_name)))
    result, all_scores = METHOD_CALLS[method_name](X, y)
    assert isinstance(result, dict)
    assert isinstance(all_scores, dict)

    if n < 2 or p == 0:
        assert result == {}
        assert all_scores == {}


@pytest.mark.parametrize(
    "call_fn",
    [_call_mutual_information, _call_anova, _call_chi_square, _call_relieff, _call_linear_svm],
)
def test_nan_injection_no_unhandled_exceptions(call_fn) -> None:
    X, y = _toy_xy(5, 20, seed=123)
    X[:, 0] = np.nan
    X[0, 1] = np.nan
    result, all_scores = call_fn(X, y)
    assert isinstance(result, dict)
    assert isinstance(all_scores, dict)


def test_copula_knockoff_short_circuits_for_small_n() -> None:
    class DummyKnockoff:
        pass

    X, y = _toy_xy(1, 10, seed=9)
    result, all_scores = copula_knockoff_selection(
        X,
        y,
        5,
        CopulaKnockoffSelectorClass=DummyKnockoff,
        copula_knockoff_draws=10,
        copula_alpha_kn=0.1,
        copula_alpha_ebh=0.2,
        copula_truncation_level=None,
        copula_generator="copula",
        copula_deepdrk_latent_fraction=0.35,
        copula_deepdrk_noise_scale=1.0,
        copula_derandomize_runs=1,
        copula_stabilizer_runs=1,
        copula_stabilizer_use_ebh=False,
        copula_stabilizer_seed_stride=997,
        random_state=0,
    )
    assert result == {}
    assert all_scores == {}


def test_preprocessing_remove_correlated_handles_empty_and_tiny_inputs() -> None:
    empty_df = pd.DataFrame(np.empty((3, 0)))
    out_empty, dropped_empty = remove_correlated_features(empty_df, threshold=0.9)
    assert out_empty.shape == empty_df.shape
    assert dropped_empty == []

    tiny_df = pd.DataFrame({"a": [1.0], "b": [2.0], "c": [3.0]})
    out_tiny, dropped_tiny = remove_correlated_features(tiny_df, threshold=0.9)
    assert list(out_tiny.columns) == ["a", "b", "c"]
    assert dropped_tiny == []


def test_transform_resets_eval_multimodel_fold_log_across_calls() -> None:
    selector = FeatureSelector()
    selector.selected_features_indices_ = np.asarray([0], dtype=int)
    X = np.ones((3, 1), dtype=float)

    selector._eval_multimodel_fold_log = [{"fold": 1}]
    _ = selector.transform(X)
    assert selector._eval_multimodel_fold_log == []

    selector._eval_multimodel_fold_log = [{"fold": 2}, {"fold": 3}]
    _ = selector.transform(X)
    assert selector._eval_multimodel_fold_log == []


def test_run_selection_methods_timeout_path_marks_method_timeout() -> None:
    selector = FeatureSelector(
        enabled_methods={"mutual_information"},
        method_timeout_seconds=0.05,
    )

    def _slow_method(*args, **kwargs):
        del args, kwargs
        time.sleep(0.2)
        return {"selected_indices": np.asarray([0], dtype=int), "scores": {0: 1.0}}, {0: 1.0}

    selector._mutual_information_selection = _slow_method  # type: ignore[method-assign]
    X, y = _toy_xy(8, 4, seed=77)
    results, _ = selector._run_selection_methods(X, y, n_target=2)
    payload = results["mutual_information"][0]
    assert bool(payload.get("timed_out", False)) is True


def test_run_selection_methods_exception_path_isolated_per_method() -> None:
    selector = FeatureSelector(
        enabled_methods={"mutual_information"},
        method_timeout_seconds=0.0,
    )

    def _raising_method(*args, **kwargs):
        del args, kwargs
        raise RuntimeError("intentional-test-error")

    selector._mutual_information_selection = _raising_method  # type: ignore[method-assign]
    X, y = _toy_xy(8, 4, seed=78)
    results, _ = selector._run_selection_methods(X, y, n_target=2)
    assert results["mutual_information"] == ({}, {})


def test_fit_transform_warns_on_hdlss_regime() -> None:
    selector = FeatureSelector(
        enabled_methods={"mutual_information"},
        n_bootstrap_iterations=1,
        inner_cv_splits=2,
        inner_cv_repeats=1,
        method_timeout_seconds=0.0,
    )
    X, y = _toy_xy(6, 300, seed=90)
    with pytest.warns(RuntimeWarning, match="HDLSS regime detected"):
        X_selected, result = selector.fit_transform(X, y, n_final_features=5, return_result_object=True)
    assert X_selected.shape[0] == 6
    assert result is not None


def test_parallel_distribution_fit_timeout_marks_failed_results(monkeypatch: pytest.MonkeyPatch) -> None:
    class DummyFuture:
        def __init__(self) -> None:
            self._cancelled = False

        def cancel(self) -> bool:
            self._cancelled = True
            return True

        def result(self, timeout=None):
            del timeout
            return None

    class DummyExecutor:
        def __init__(self, *args, **kwargs) -> None:
            del args, kwargs

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc, tb):
            del exc_type, exc, tb
            return False

        def submit(self, fn, args):
            del fn, args
            return DummyFuture()

    def _as_completed_timeout(_futures, timeout=None):
        del _futures, timeout
        raise uds_mod.FuturesTimeoutError()

    monkeypatch.setattr(uds_mod, "ProcessPoolExecutor", DummyExecutor)
    monkeypatch.setattr(uds_mod, "as_completed", _as_completed_timeout)

    selector = uds_mod.UnifiedDistributionSelectorV6(n_jobs=2)
    data = np.linspace(0.1, 1.0, num=16)
    dist_name = next(iter(selector.distributions.keys()))
    results = selector._fit_distributions_parallel(
        data,
        features=None,
        dist_names=[dist_name],
        verbose=False,
    )
    assert len(results) == 1
    assert bool(results[0].success) is False
    assert "timeout" in str(results[0].error).lower()



"""Tests for multi-classifier evaluation proxy (T-001).

Covers:
- Toggle OFF regression (bit-for-bit identical to single-model)
- Toggle ON determinism (fixed seed → same score)
- Aggregation math correctness (mean, min, CVaR)
- Single-class edge cases
- Per-model scores returned
- Model library validity
- LinearSVC predict consistency (no predict_proba fallback)
- Config integration
- Pinned-seed regression
"""

import numpy as np
import pytest

from tabnetics.feature_selection.cv import (
    CVEvaluationContext,
    FoldLeakageError,
    fit_and_score_fold,
    fit_and_score_fold_multimodel,
    get_inner_cv_splits,
    _get_eval_model,
    _score_fitted_model,
    resolve_performance_weights,
    safe_balanced_accuracy,
)
from tabnetics.feature_selection.config import (
    EvaluationConfig,
    FeatureSelectorConfig,
)


# ── Fixtures ──────────────────────────────────────────────────────────

def _make_binary_data(n=80, p=6, seed=42):
    """Synthetic binary classification dataset."""
    rng = np.random.default_rng(seed)
    X = rng.standard_normal((n, p))
    y = (X[:, 0] + 0.5 * X[:, 1] > 0).astype(int)
    mid = n // 2
    return X[:mid], y[:mid], X[mid:], y[mid:]


def _make_multiclass_data(n=120, p=6, n_classes=4, seed=42):
    """Synthetic multiclass classification dataset."""
    rng = np.random.default_rng(seed)
    X = rng.standard_normal((n, p))
    y = rng.integers(0, n_classes, size=n)
    mid = n // 2
    return X[:mid], y[:mid], X[mid:], y[mid:]


def _make_regression_data(n=80, p=6, seed=42):
    """Synthetic regression dataset."""
    rng = np.random.default_rng(seed)
    X = rng.standard_normal((n, p))
    y = X[:, 0] * 2.0 + rng.standard_normal(n) * 0.3
    mid = n // 2
    return X[:mid], y[:mid], X[mid:], y[mid:]


# Default performance weights (matching FeatureSelector defaults)
_PERF_KWARGS = dict(
    performance_balanced_weight=0.6,
    performance_macro_f1_weight=0.4,
    performance_use_adaptive_imbalance=False,
    performance_imbalance_ratio_trigger=1.75,
    performance_min_classes_for_adaptive=3,
)


# ── 1. Toggle OFF regression: bit-for-bit identical ──────────────────

class TestToggleOffRegression:
    """When eval_models_enabled=False, behavior must be bit-for-bit identical."""

    def test_binary_bit_for_bit(self):
        X_tr, y_tr, X_val, y_val = _make_binary_data()
        score_single, signal_single = fit_and_score_fold(
            X_tr, y_tr, X_val, y_val, 'classification', 42, **_PERF_KWARGS,
        )
        # Multi-model with only lr_l2 and mean agg should reproduce single LR
        score_multi, signal_multi, per_model = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2",), eval_aggregate="mean", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        assert score_single == pytest.approx(score_multi, abs=1e-12), \
            f"Single={score_single}, Multi={score_multi}"
        np.testing.assert_array_equal(signal_single, signal_multi)

    def test_multiclass_bit_for_bit(self):
        X_tr, y_tr, X_val, y_val = _make_multiclass_data()
        score_single, signal_single = fit_and_score_fold(
            X_tr, y_tr, X_val, y_val, 'classification', 42, **_PERF_KWARGS,
        )
        score_multi, signal_multi, _ = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2",), eval_aggregate="mean", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        assert score_single == pytest.approx(score_multi, abs=1e-12)
        np.testing.assert_array_equal(signal_single, signal_multi)

    def test_regression_fallback(self):
        X_tr, y_tr, X_val, y_val = _make_regression_data()
        score_single, signal_single = fit_and_score_fold(
            X_tr, y_tr, X_val, y_val, 'regression', 42, **_PERF_KWARGS,
        )
        score_multi, signal_multi, per_model = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'regression', 42,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="mean", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        assert score_single == pytest.approx(score_multi, abs=1e-12)
        np.testing.assert_array_equal(signal_single, signal_multi)
        assert "_fallback_regression" in per_model


# ── 2. Toggle ON determinism ─────────────────────────────────────────

class TestToggleOnDeterminism:
    """Same seed → same multi-model score."""

    def test_deterministic_binary(self):
        X_tr, y_tr, X_val, y_val = _make_binary_data()
        results = []
        for _ in range(3):
            s, sig, pm = fit_and_score_fold_multimodel(
                X_tr, y_tr, X_val, y_val, 'classification', 42,
                eval_models=("lr_l2", "linear_svc", "rf_small"),
                eval_aggregate="mean", eval_cvar_alpha=0.33,
                **_PERF_KWARGS,
            )
            results.append((s, sig.copy(), dict(pm)))
        for i in range(1, len(results)):
            assert results[0][0] == pytest.approx(results[i][0], abs=1e-12)
            np.testing.assert_array_equal(results[0][1], results[i][1])

    def test_deterministic_multiclass(self):
        X_tr, y_tr, X_val, y_val = _make_multiclass_data()
        s1, _, pm1 = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 99,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="mean", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        s2, _, pm2 = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 99,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="mean", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        assert s1 == pytest.approx(s2, abs=1e-12)
        for key in pm1:
            assert pm1[key] == pytest.approx(pm2[key], abs=1e-12)


# ── 3. Aggregation math correctness ──────────────────────────────────

class TestAggregationMath:
    """Verify mean, min, CVaR produce correct results."""

    def test_mean_aggregation(self):
        X_tr, y_tr, X_val, y_val = _make_binary_data()
        _, _, pm = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="mean", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        scores = np.array([pm[k] for k in ("lr_l2", "linear_svc", "rf_small")])
        expected = float(np.mean(scores))
        actual, _, _ = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="mean", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        assert actual == pytest.approx(expected, abs=1e-12)

    def test_min_aggregation(self):
        X_tr, y_tr, X_val, y_val = _make_binary_data()
        # First get per-model scores with mean to know what they are
        _, _, pm = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="mean", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        scores = np.array([pm[k] for k in ("lr_l2", "linear_svc", "rf_small")])
        expected_min = float(np.min(scores))
        actual_min, _, _ = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="min", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        assert actual_min == pytest.approx(expected_min, abs=1e-12)

    def test_cvar_aggregation(self):
        X_tr, y_tr, X_val, y_val = _make_binary_data()
        _, _, pm = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="mean", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        scores = np.array([pm[k] for k in ("lr_l2", "linear_svc", "rf_small")])
        # CVaR(alpha=0.5): mean of bottom ceil(0.5*3)=2 scores
        alpha = 0.5
        k = max(1, int(np.ceil(alpha * len(scores))))
        sorted_scores = np.sort(scores)
        expected_cvar = float(np.mean(sorted_scores[:k]))

        actual_cvar, _, _ = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="cvar", eval_cvar_alpha=0.5,
            **_PERF_KWARGS,
        )
        assert actual_cvar == pytest.approx(expected_cvar, abs=1e-12)

    def test_cvar_alpha_one_equals_mean(self):
        """CVaR with alpha=1.0 is mean of all scores."""
        X_tr, y_tr, X_val, y_val = _make_binary_data()
        mean_score, _, _ = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="mean", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        cvar_score, _, _ = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="cvar", eval_cvar_alpha=1.0,
            **_PERF_KWARGS,
        )
        assert cvar_score == pytest.approx(mean_score, abs=1e-12)

    def test_cvar_alpha_small_equals_min(self):
        """CVaR with very small alpha ≈ min (bottom 1 score)."""
        X_tr, y_tr, X_val, y_val = _make_binary_data()
        min_score, _, _ = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="min", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        cvar_score, _, _ = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="cvar", eval_cvar_alpha=0.01,
            **_PERF_KWARGS,
        )
        # alpha=0.01 → k=ceil(0.01*3)=1 → mean of bottom 1 = min
        assert cvar_score == pytest.approx(min_score, abs=1e-12)


# ── 4. Single-class edge case ────────────────────────────────────────

class TestSingleClassEdge:
    """All training data is one class."""

    def test_single_class_train(self):
        rng = np.random.default_rng(42)
        X_tr = rng.standard_normal((30, 4))
        y_tr = np.zeros(30, dtype=int)  # single class
        X_val = rng.standard_normal((10, 4))
        y_val = np.array([0, 0, 0, 1, 1, 1, 0, 0, 1, 0])

        score, signal, pm = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="mean", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        assert np.isfinite(score)
        assert signal.shape == (10,)
        assert "_single_class" in pm

    def test_empty_class_in_train(self):
        """Only one class present in y_train."""
        rng = np.random.default_rng(7)
        X_tr = rng.standard_normal((20, 3))
        y_tr = np.ones(20, dtype=int)  # all class 1
        X_val = rng.standard_normal((5, 3))
        y_val = np.array([0, 1, 1, 0, 1])

        score, signal, pm = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="mean", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        assert np.isfinite(score)
        assert signal.shape == (5,)


# ── 5. Per-model scores returned ─────────────────────────────────────

class TestPerModelScores:
    """Verify diagnostics dict has per-model entries."""

    def test_all_models_in_diagnostics(self):
        X_tr, y_tr, X_val, y_val = _make_binary_data()
        _, _, pm = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="mean", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        assert "lr_l2" in pm
        assert "linear_svc" in pm
        assert "rf_small" in pm
        for v in pm.values():
            assert isinstance(v, float)
            assert np.isfinite(v)


class TestFoldHygieneAndFailureModes:
    """Fold-hygiene guard and evaluation-failure observability contracts."""

    def test_blocks_learned_aggregation_in_evaluation_fold(self):
        X_tr, y_tr, X_val, y_val = _make_binary_data()
        with pytest.raises(FoldLeakageError):
            fit_and_score_fold_multimodel(
                X_tr, y_tr, X_val, y_val, 'classification', 42,
                eval_models=("lr_l2", "linear_svc", "rf_small"),
                eval_aggregate="mean", eval_cvar_alpha=0.33,
                eval_model_weight_strategy="learned",
                evaluation_context=CVEvaluationContext(
                    purpose="evaluation_fold",
                    allow_learned_model_aggregation=False,
                ),
                **_PERF_KWARGS,
            )

    def test_allows_explicit_override_when_context_permits(self):
        X_tr, y_tr, X_val, y_val = _make_binary_data()
        score, signal, pm = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2",),
            eval_aggregate="mean", eval_cvar_alpha=0.33,
            eval_model_weight_strategy="learned",
            evaluation_context=CVEvaluationContext(
                purpose="evaluation_fold",
                allow_learned_model_aggregation=True,
            ),
            **_PERF_KWARGS,
        )
        assert np.isfinite(score)
        assert signal.shape == (X_val.shape[0],)
        assert "lr_l2" in pm

    def test_non_strict_mode_tracks_failures_and_aggregates_model_scores_only(self, monkeypatch):
        X_tr, y_tr, X_val, y_val = _make_binary_data()

        def _boom(*args, **kwargs):
            raise RuntimeError("forced eval failure")

        monkeypatch.setattr("tabnetics.feature_selection.cv._score_fitted_model", _boom)
        score, _, pm = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2", "linear_svc"),
            eval_aggregate="mean", eval_cvar_alpha=0.33,
            eval_failure_strict_mode=False,
            **_PERF_KWARGS,
        )
        assert pm["_evaluation_failures_total"] == 2.0
        assert pm["_evaluation_failures_lr_l2"] == 1.0
        assert pm["_evaluation_failures_linear_svc"] == 1.0
        assert score == pytest.approx(0.0, abs=1e-12)

    def test_strict_mode_raises_on_eval_failure(self, monkeypatch):
        X_tr, y_tr, X_val, y_val = _make_binary_data()

        def _boom(*args, **kwargs):
            raise RuntimeError("forced eval failure")

        monkeypatch.setattr("tabnetics.feature_selection.cv._score_fitted_model", _boom)
        with pytest.raises(RuntimeError, match="forced eval failure"):
            fit_and_score_fold_multimodel(
                X_tr, y_tr, X_val, y_val, 'classification', 42,
                eval_models=("lr_l2", "linear_svc"),
                eval_aggregate="mean", eval_cvar_alpha=0.33,
                eval_failure_strict_mode=True,
                **_PERF_KWARGS,
            )

    def test_subset_models(self):
        X_tr, y_tr, X_val, y_val = _make_binary_data()
        _, _, pm = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2", "rf_small"),
            eval_aggregate="mean", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        assert "lr_l2" in pm
        assert "rf_small" in pm
        assert "linear_svc" not in pm

def test_get_inner_cv_splits_handles_n_lt_2():
    X = np.zeros((1, 4), dtype=float)
    y = np.array([0], dtype=int)
    splits = get_inner_cv_splits(
        X,
        y,
        "classification",
        random_state=42,
        inner_cv_splits=5,
        inner_cv_repeats=3,
    )
    assert splits == []


# ── 6. Model library validity ────────────────────────────────────────

class TestModelLibrary:
    """All 3 model keys resolve to valid classifiers."""

    @pytest.mark.parametrize("key", ["lr_l2", "linear_svc", "rf_small"])
    def test_model_instantiation(self, key):
        model = _get_eval_model(key, random_state=42, max_iter=2000)
        assert hasattr(model, 'fit')
        assert hasattr(model, 'predict')

    def test_unknown_model_raises(self):
        with pytest.raises(ValueError, match="Unknown eval model"):
            _get_eval_model("xgboost_mega", random_state=42)

    @pytest.mark.parametrize("key", ["lr_l2", "linear_svc", "rf_small"])
    def test_model_can_fit_and_predict(self, key):
        rng = np.random.default_rng(42)
        X = rng.standard_normal((40, 4))
        y = (X[:, 0] > 0).astype(int)
        model = _get_eval_model(key, random_state=42)
        model.fit(X, y)
        preds = model.predict(X)
        assert preds.shape == (40,)


# ── 7. LinearSVC predict consistency ─────────────────────────────────

class TestLinearSVCFallback:
    """LinearSVC doesn't have predict_proba — verify signal uses predictions."""

    def test_no_predict_proba(self):
        model = _get_eval_model("linear_svc", random_state=42)
        assert not hasattr(model, "predict_proba")

    def test_score_without_proba(self):
        X_tr, y_tr, X_val, y_val = _make_binary_data()
        model = _get_eval_model("linear_svc", random_state=42)
        model.fit(X_tr, y_tr)
        w_bal, w_f1 = 0.6, 0.4
        score, signal = _score_fitted_model(model, X_val, y_val, w_bal, w_f1)
        assert np.isfinite(score)
        # Signal should be integer predictions (no proba)
        assert signal.shape == y_val.shape
        # All signal values should be class labels (0 or 1)
        unique_signal = np.unique(signal)
        assert all(v in {0.0, 1.0} for v in unique_signal)


# ── 8. Config integration ────────────────────────────────────────────

class TestConfigIntegration:
    """EvaluationConfig defaults produce correct behavior."""

    def test_default_config(self):
        cfg = EvaluationConfig()
        assert cfg.eval_models_enabled is False
        assert cfg.eval_models == ("lr_l2", "linear_svc", "rf_small")
        assert cfg.eval_aggregate == "mean"
        assert cfg.eval_cvar_alpha == pytest.approx(0.33)

    def test_config_in_feature_selector_config(self):
        fsc = FeatureSelectorConfig()
        assert hasattr(fsc, 'evaluation')
        assert isinstance(fsc.evaluation, EvaluationConfig)
        assert fsc.evaluation.eval_models_enabled is False

    def test_custom_config(self):
        cfg = EvaluationConfig(
            eval_models_enabled=True,
            eval_models=("lr_l2", "rf_small"),
            eval_aggregate="cvar",
            eval_cvar_alpha=0.5,
        )
        assert cfg.eval_models_enabled is True
        assert cfg.eval_models == ("lr_l2", "rf_small")
        assert cfg.eval_aggregate == "cvar"
        assert cfg.eval_cvar_alpha == 0.5

    def test_from_config_wiring(self):
        """FeatureSelector.from_config passes evaluation config through."""
        fsc = FeatureSelectorConfig(
            evaluation=EvaluationConfig(eval_models_enabled=True, eval_aggregate="min"),
        )
        from tabnetics.feature_selection.base import FeatureSelector
        fs = FeatureSelector.from_config(fsc)
        assert fs.eval_models_enabled is True
        assert fs.eval_aggregate == "min"

    def test_from_config_default_off(self):
        """Default config → eval_models_enabled=False on FeatureSelector."""
        fsc = FeatureSelectorConfig()
        from tabnetics.feature_selection.base import FeatureSelector
        fs = FeatureSelector.from_config(fsc)
        assert fs.eval_models_enabled is False


# ── 9. Pinned-seed regression test ───────────────────────────────────

class TestPinnedSeedRegression:
    """Multi-model produces a stable score for fixed synthetic data."""

    def test_pinned_binary_multimodel(self):
        """Pinned regression: binary classification with all 3 models."""
        rng = np.random.default_rng(12345)
        X = rng.standard_normal((100, 5))
        y = (X[:, 0] + 0.8 * X[:, 2] > 0).astype(int)
        X_tr, y_tr = X[:60], y[:60]
        X_val, y_val = X[60:], y[60:]

        score, signal, pm = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="mean", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        # Score must be finite and in reasonable range
        assert np.isfinite(score)
        assert 0.0 <= score <= 1.5  # AUC bonus can push slightly above 1.0
        assert signal.shape == (40,)
        model_keys = ("lr_l2", "linear_svc", "rf_small")
        assert all(k in pm for k in model_keys)
        assert pm["_evaluation_failures_total"] == 0.0
        # All per-model scores finite
        for k, v in pm.items():
            assert np.isfinite(v), f"Model {k} score is not finite: {v}"

    def test_pinned_multiclass_multimodel(self):
        """Pinned regression: multiclass classification with all 3 models."""
        rng = np.random.default_rng(54321)
        X = rng.standard_normal((120, 6))
        y = (np.digitize(X[:, 0], bins=[-0.5, 0.5, 1.5])).astype(int)
        X_tr, y_tr = X[:80], y[:80]
        X_val, y_val = X[80:], y[80:]

        score, signal, pm = fit_and_score_fold_multimodel(
            X_tr, y_tr, X_val, y_val, 'classification', 42,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="mean", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        assert np.isfinite(score)
        assert signal.shape == (40,)
        model_keys = ("lr_l2", "linear_svc", "rf_small")
        assert all(k in pm for k in model_keys)
        assert pm["_evaluation_failures_total"] == 0.0


# ── 10. No fold leakage: sanity check ────────────────────────────────

class TestNoFoldLeakage:
    """Each model is fitted on training fold only, scored on val fold only."""

    def test_score_changes_with_different_val(self):
        """Different validation data → different scores (no data contamination)."""
        rng = np.random.default_rng(42)
        X = rng.standard_normal((100, 5))
        y = (X[:, 0] > 0).astype(int)
        X_tr, y_tr = X[:60], y[:60]

        score_a, _, _ = fit_and_score_fold_multimodel(
            X_tr, y_tr, X[60:80], y[60:80], 'classification', 42,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="mean", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        score_b, _, _ = fit_and_score_fold_multimodel(
            X_tr, y_tr, X[80:100], y[80:100], 'classification', 42,
            eval_models=("lr_l2", "linear_svc", "rf_small"),
            eval_aggregate="mean", eval_cvar_alpha=0.33,
            **_PERF_KWARGS,
        )
        # Scores should differ with different validation data
        # (unless by extreme coincidence, which is astronomically unlikely)
        assert score_a != pytest.approx(score_b, abs=1e-6)


# ── 11. Base.py integration ──────────────────────────────────────────

class TestBaseIntegration:
    """Test _fit_and_score_fold dispatching on FeatureSelector."""

    def test_toggle_off_uses_single_model(self):
        from tabnetics.feature_selection.base import FeatureSelector
        fs = FeatureSelector(random_state=42, eval_models_enabled=False)
        X_tr, y_tr, X_val, y_val = _make_binary_data()
        score, signal = fs._fit_and_score_fold(X_tr, y_tr, X_val, y_val)
        assert np.isfinite(score)
        # No multi-model diagnostics logged
        assert len(fs._eval_multimodel_fold_log) == 0

    def test_toggle_on_uses_multimodel(self):
        from tabnetics.feature_selection.base import FeatureSelector
        fs = FeatureSelector(random_state=42, eval_models_enabled=True)
        X_tr, y_tr, X_val, y_val = _make_binary_data()
        score, signal = fs._fit_and_score_fold(X_tr, y_tr, X_val, y_val)
        assert np.isfinite(score)
        # Multi-model diagnostics should be logged
        assert len(fs._eval_multimodel_fold_log) == 1
        log_entry = fs._eval_multimodel_fold_log[0]
        assert "lr_l2" in log_entry
        assert "linear_svc" in log_entry
        assert "rf_small" in log_entry

    def test_toggle_on_deterministic_via_base(self):
        from tabnetics.feature_selection.base import FeatureSelector
        X_tr, y_tr, X_val, y_val = _make_binary_data()
        fs1 = FeatureSelector(random_state=42, eval_models_enabled=True)
        s1, sig1 = fs1._fit_and_score_fold(X_tr, y_tr, X_val, y_val)
        fs2 = FeatureSelector(random_state=42, eval_models_enabled=True)
        s2, sig2 = fs2._fit_and_score_fold(X_tr, y_tr, X_val, y_val)
        assert s1 == pytest.approx(s2, abs=1e-12)
        np.testing.assert_array_equal(sig1, sig2)



"""Tests for TabPFN benchmark classifier wrapper (VAL12_Suggestions §4.2)."""

import numpy as np
import pytest
from unittest.mock import patch, MagicMock

from tabnetics.feature_selection.methods.tabpfn_classifier import (
    TabPFNBenchmarkClassifier,
    TABPFN_AVAILABLE,
    MAX_FEATURES,
    MAX_SAMPLES,
)


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def _make_binary_data(n=80, p=10, seed=42):
    rng = np.random.RandomState(seed)
    X = rng.randn(n, p)
    y = np.repeat([0, 1], n // 2)
    X[:n // 2, :2] += 1.5
    return X, y


def _make_multiclass_data(n=90, p=8, seed=42):
    rng = np.random.RandomState(seed)
    X = rng.randn(n, p)
    y = np.tile([0, 1, 2], n // 3)[:n]
    X[y == 0, :2] += 2.0
    X[y == 1, 2:4] += 2.0
    return X, y


# ---------------------------------------------------------------------------
# Fallback tests (always run, no tabpfn needed)
# ---------------------------------------------------------------------------


class TestTabPFNFallback:
    """Verify fallback behavior when TabPFN is unavailable or limits exceeded."""

    def test_fallback_when_too_many_features(self):
        X, y = _make_binary_data(n=50, p=MAX_FEATURES + 10)
        clf = TabPFNBenchmarkClassifier(random_state=42)
        with patch(
            "tabnetics.feature_selection.methods.tabpfn_classifier.TABPFN_AVAILABLE",
            True,
        ):
            with pytest.warns(RuntimeWarning, match="n_features="):
                clf.fit(X, y)
        assert clf.exceeded_limits_
        assert clf.fallback_reason_ is not None
        assert "n_features" in clf.fallback_reason_
        assert not clf.is_using_tabpfn

    def test_fallback_when_too_many_samples(self):
        X, y = _make_binary_data(n=MAX_SAMPLES + 100, p=5)
        clf = TabPFNBenchmarkClassifier(random_state=42)
        with patch(
            "tabnetics.feature_selection.methods.tabpfn_classifier.TABPFN_AVAILABLE",
            True,
        ):
            with pytest.warns(RuntimeWarning, match="n_samples="):
                clf.fit(X, y)
        assert clf.exceeded_limits_
        assert "n_samples" in clf.fallback_reason_
        assert not clf.is_using_tabpfn

    def test_fallback_predict_produces_valid_labels(self):
        X, y = _make_binary_data(n=50, p=MAX_FEATURES + 1)
        clf = TabPFNBenchmarkClassifier(random_state=42)
        with pytest.warns(RuntimeWarning):
            clf.fit(X, y)
        preds = clf.predict(X)
        assert preds.shape == (50,)
        assert set(preds).issubset({0, 1})

    def test_fallback_predict_proba_shape(self):
        X, y = _make_binary_data(n=50, p=MAX_FEATURES + 1)
        clf = TabPFNBenchmarkClassifier(random_state=42)
        with pytest.warns(RuntimeWarning):
            clf.fit(X, y)
        probs = clf.predict_proba(X)
        assert probs.shape == (50, 2)
        np.testing.assert_allclose(probs.sum(axis=1), 1.0, atol=1e-10)

    def test_custom_limits(self):
        X, y = _make_binary_data(n=50, p=20)
        clf = TabPFNBenchmarkClassifier(max_features=15, max_samples=40, random_state=0)
        with patch(
            "tabnetics.feature_selection.methods.tabpfn_classifier.TABPFN_AVAILABLE",
            True,
        ):
            with pytest.warns(RuntimeWarning, match="n_features="):
                clf.fit(X, y)
        assert clf.exceeded_limits_

    def test_custom_sample_limit(self):
        X, y = _make_binary_data(n=50, p=5)
        clf = TabPFNBenchmarkClassifier(max_samples=30, random_state=0)
        with patch(
            "tabnetics.feature_selection.methods.tabpfn_classifier.TABPFN_AVAILABLE",
            True,
        ):
            with pytest.warns(RuntimeWarning, match="n_samples="):
                clf.fit(X, y)
        assert clf.exceeded_limits_

    def test_unfitted_predict_raises(self):
        clf = TabPFNBenchmarkClassifier()
        with pytest.raises(Exception):  # NotFittedError or AttributeError
            clf.predict(np.zeros((5, 3)))

    def test_multiclass_fallback(self):
        X, y = _make_multiclass_data(n=60, p=MAX_FEATURES + 5)
        clf = TabPFNBenchmarkClassifier(random_state=42)
        with pytest.warns(RuntimeWarning):
            clf.fit(X, y)
        preds = clf.predict(X)
        assert preds.shape == (60,)
        probs = clf.predict_proba(X)
        assert probs.shape == (60, 3)

    def test_fallback_when_package_unavailable(self):
        """Simulate tabpfn not installed."""
        with patch(
            "tabnetics.feature_selection.methods.tabpfn_classifier.TABPFN_AVAILABLE",
            False,
        ):
            X, y = _make_binary_data(n=50, p=5)
            clf = TabPFNBenchmarkClassifier(random_state=42)
            with pytest.warns(RuntimeWarning, match="not installed"):
                clf.fit(X, y)
            assert clf.exceeded_limits_
            assert "not installed" in clf.fallback_reason_

    def test_classes_attribute_set(self):
        X, y = _make_multiclass_data(n=60, p=MAX_FEATURES + 1)
        clf = TabPFNBenchmarkClassifier(random_state=42)
        with pytest.warns(RuntimeWarning):
            clf.fit(X, y)
        np.testing.assert_array_equal(clf.classes_, [0, 1, 2])

    def test_n_features_in_set(self):
        X, y = _make_binary_data(n=50, p=MAX_FEATURES + 1)
        clf = TabPFNBenchmarkClassifier(random_state=42)
        with pytest.warns(RuntimeWarning):
            clf.fit(X, y)
        assert clf.n_features_in_ == MAX_FEATURES + 1


# ---------------------------------------------------------------------------
# Tests that use real TabPFN (skip if not installed)
# ---------------------------------------------------------------------------


@pytest.mark.skipif(not TABPFN_AVAILABLE, reason="tabpfn not installed")
class TestTabPFNReal:
    """Verify real TabPFN integration (requires tabpfn package)."""

    def test_fit_predict_binary(self):
        X, y = _make_binary_data(n=80, p=10)
        clf = TabPFNBenchmarkClassifier(random_state=42)
        clf.fit(X[:60], y[:60])
        assert clf.is_using_tabpfn
        preds = clf.predict(X[60:])
        assert preds.shape == (20,)
        assert set(preds).issubset({0, 1})

    def test_predict_proba_binary(self):
        X, y = _make_binary_data(n=80, p=10)
        clf = TabPFNBenchmarkClassifier(random_state=42)
        clf.fit(X[:60], y[:60])
        probs = clf.predict_proba(X[60:])
        assert probs.shape == (20, 2)
        np.testing.assert_allclose(probs.sum(axis=1), 1.0, atol=1e-6)

    def test_fit_predict_multiclass(self):
        X, y = _make_multiclass_data(n=90, p=8)
        clf = TabPFNBenchmarkClassifier(random_state=42)
        clf.fit(X[:60], y[:60])
        preds = clf.predict(X[60:])
        assert preds.shape == (30,)
        probs = clf.predict_proba(X[60:])
        assert probs.shape == (30, 3)

    def test_within_limits_uses_tabpfn(self):
        X, y = _make_binary_data(n=50, p=10)
        clf = TabPFNBenchmarkClassifier(random_state=42)
        clf.fit(X, y)
        assert clf.is_using_tabpfn
        assert not clf.exceeded_limits_

    def test_at_feature_boundary(self):
        rng = np.random.RandomState(42)
        X = rng.randn(50, MAX_FEATURES)
        y = np.repeat([0, 1], 25)
        clf = TabPFNBenchmarkClassifier(random_state=42)
        clf.fit(X, y)
        assert clf.is_using_tabpfn

    def test_at_sample_boundary(self):
        rng = np.random.RandomState(42)
        X = rng.randn(MAX_SAMPLES, 10)
        y = np.repeat([0, 1], MAX_SAMPLES // 2)
        clf = TabPFNBenchmarkClassifier(random_state=42)
        clf.fit(X, y)
        assert clf.is_using_tabpfn


# ---------------------------------------------------------------------------
# Config toggle test
# ---------------------------------------------------------------------------


class TestBenchmarkConfig:
    """Verify benchmark_tabpfn_enabled config field exists."""

    def test_config_field_default_false(self):
        from tabnetics.pipeline.pipeline import DFFSConfig
        cfg = DFFSConfig()
        assert hasattr(cfg, "benchmark_tabpfn_enabled")
        assert cfg.benchmark_tabpfn_enabled is False



"""Dataset meta-feature extraction helpers."""

from __future__ import annotations

from typing import Dict

import numpy as np
from scipy.optimize import curve_fit
from scipy.stats import entropy as _shannon_entropy


def extract_meta_features(X: np.ndarray, y: np.ndarray) -> Dict[str, float]:
    """Extract dataset meta-features useful for tier assignment and analysis."""
    X = np.asarray(X, dtype=float)
    y = np.asarray(y)

    n, p = X.shape if X.ndim == 2 else (X.shape[0], 1)
    if X.ndim == 1:
        X = X.reshape(-1, 1)

    classes, counts = np.unique(y, return_counts=True)
    class_count = float(len(classes))
    if class_count > 1:
        proportions = counts / counts.sum()
        raw_entropy = _shannon_entropy(proportions, base=np.e)
        class_balance_entropy = float(raw_entropy / np.log(class_count))
    else:
        class_balance_entropy = 0.0

    p_over_n = float(p) / float(n) if n > 0 else 0.0

    if p >= 3 and n >= 2:
        rng = np.random.RandomState(42)
        max_cols = 200
        if p > max_cols:
            col_idx = rng.choice(p, max_cols, replace=False)
            X_sub = X[:, col_idx]
        else:
            X_sub = X
        with np.errstate(divide="ignore", invalid="ignore"):
            corr = np.corrcoef(X_sub, rowvar=False)
        corr = np.nan_to_num(corr, nan=0.0)
        tri_idx = np.triu_indices_from(corr, k=1)
        abs_corr = np.sort(np.abs(corr[tri_idx]))[::-1]
        xs = np.arange(len(abs_corr), dtype=float)
        try:
            def _exp_decay(x, a, b):
                return a * np.exp(-b * x)

            popt, _ = curve_fit(
                _exp_decay,
                xs,
                abs_corr,
                p0=[abs_corr[0] if len(abs_corr) else 1.0, 0.01],
                maxfev=2000,
            )
            correlation_spectrum_decay = float(max(popt[1], 0.0))
        except Exception:
            correlation_spectrum_decay = 0.0
    else:
        correlation_spectrum_decay = 0.0

    if p > 0:
        heaping_count = 0
        for j in range(p):
            col = X[:, j]
            valid = col[~np.isnan(col)]
            if len(valid) > 0 and np.all(valid == np.round(valid)):
                heaping_count += 1
        heaping_fraction = float(heaping_count) / float(p)
    else:
        heaping_fraction = 0.0

    return {
        "n": float(n),
        "p": float(p),
        "p_over_n": float(p_over_n),
        "class_count": float(class_count),
        "class_balance_entropy": float(class_balance_entropy),
        "correlation_spectrum_decay": float(correlation_spectrum_decay),
        "heaping_fraction": float(heaping_fraction),
    }


__all__ = ["extract_meta_features"]
